"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os

import salt.modules.solr as solr
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SolrTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.solr
    """

    def setup_loader_modules(self):
        return {solr: {}}

    def test_lucene_version(self):
        """
        Test to get the lucene version that solr is using.
        """
        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(
                solr, "_get_none_or_value", side_effect=[None, True, True]
            ):
                with patch.object(
                    solr, "_check_for_cores", side_effect=[True, False, False]
                ):
                    tempdict = {
                        "success": "success",
                        "errors": "errors",
                        "data": {"lucene": {"lucene-spec-version": 1}},
                    }
                    with patch.object(
                        solr,
                        "_get_admin_info",
                        side_effect=[tempdict, tempdict, {"success": None}],
                    ):
                        with patch.dict(
                            solr.__salt__,
                            {"config.option": MagicMock(return_value=["A"])},
                        ):

                            with patch.object(
                                solr, "_update_return_dict", return_value={"A": "a"}
                            ):
                                self.assertDictEqual(
                                    solr.lucene_version("c"), {"A": "a"}
                                )

                            self.assertDictEqual(solr.lucene_version("c"), {"A": "a"})

                            self.assertDictEqual(
                                solr.lucene_version("c"), {"success": None}
                            )

    def test_version(self):
        """
        Test to get the solr version for the core specified
        """
        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(
                solr, "_get_none_or_value", side_effect=[None, True, True]
            ):
                with patch.object(
                    solr, "_check_for_cores", side_effect=[True, False, False]
                ):

                    tempdict = {
                        "success": "success",
                        "errors": "errors",
                        "warnings": "warnings",
                        "data": {"lucene": {"solr-spec-version": 1}},
                    }
                    with patch.object(
                        solr, "_get_admin_info", side_effect=[tempdict, tempdict]
                    ):
                        with patch.dict(
                            solr.__opts__, {"solr.cores": MagicMock(return_value=["A"])}
                        ):
                            with patch.object(
                                solr, "_update_return_dict", return_value={"A": "a"}
                            ):
                                self.assertDictEqual(solr.version(), {"A": "a"})

                            self.assertDictEqual(solr.version(), {"A": "a"})

                    with patch.object(
                        solr, "_get_admin_info", return_value={"success": None}
                    ):
                        self.assertDictEqual(solr.version(), {"success": None})

    def test_optimize(self):
        """
        Test to search queries fast, but it is a very expensive operation.
        """
        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(solr, "_get_none_or_value", side_effect=[None, True]):
                with patch.object(solr, "_check_for_cores", side_effect=[True, False]):
                    tempdict = {
                        "success": "success",
                        "errors": "errors",
                        "warnings": "warnings",
                        "data": {"lucene": {"solr-spec-version": 1}},
                    }
                    with patch.object(solr, "_format_url", return_value="A"):
                        with patch.dict(
                            solr.__salt__,
                            {"config.option": MagicMock(return_value=["A"])},
                        ):
                            with patch.object(
                                solr, "_http_request", return_value=tempdict
                            ):
                                with patch.object(
                                    solr, "_update_return_dict", return_value={"A": "a"}
                                ):
                                    self.assertDictEqual(solr.optimize(), {"A": "a"})

                        with patch.object(solr, "_http_request", return_value="A"):
                            self.assertEqual(solr.optimize(), "A")

    def test_ping(self):
        """
        Test to check on solr, makes sure solr can talk to the
        indexes.
        """
        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(solr, "_get_none_or_value", side_effect=[None, True]):
                with patch.object(solr, "_check_for_cores", side_effect=[True, False]):

                    tempdict = {
                        "success": "success",
                        "errors": "errors",
                        "warnings": "warnings",
                        "data": {"lucene": {"solr-spec-version": 1}},
                    }

                    with patch.dict(
                        solr.__opts__, {"solr.cores": MagicMock(return_value=["A"])}
                    ):
                        with patch.object(
                            solr, "_get_admin_info", return_value=tempdict
                        ):
                            with patch.object(
                                solr, "_update_return_dict", return_value={"A": "a"}
                            ):
                                self.assertDictEqual(solr.ping(), {"A": "a"})

                    with patch.object(solr, "_get_admin_info", return_value="A"):
                        self.assertEqual(solr.ping(), "A")

    def test_is_replication_enabled(self):
        """
        Test to check for errors, and determine if a slave
        is replicating or not.
        """
        error = 'Only "slave" minions can run "is_replication_enabled"'
        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(solr, "_is_master", side_effect=[True, False]):
                self.assertIsNone(solr.is_replication_enabled())
                with patch.object(solr, "_get_none_or_value", return_value=None):
                    with patch.object(solr, "_check_for_cores", return_value=True):
                        with patch.dict(
                            solr.__opts__, {"solr.cores": MagicMock(return_value="A")}
                        ):
                            with patch.object(
                                solr, "_replication_request", return_value="A"
                            ):
                                self.assertDictEqual(
                                    solr.is_replication_enabled(),
                                    {"A": "a", "errors": [error], "success": False},
                                )

    def test_match_index_versions(self):
        """
        Test to verifies that the master and the slave versions are in sync by
        comparing the index version.
        """
        err = 'solr.match_index_versions can only be called by "slave" minions'
        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(solr, "_is_master", side_effect=[True, False]):

                self.assertIsNone(solr.match_index_versions())

                with patch.object(solr, "_get_none_or_value", return_value=None):
                    with patch.object(solr, "_check_for_cores", return_value=True):
                        with patch.dict(
                            solr.__opts__, {"solr.cores": MagicMock(return_value="A")}
                        ):
                            with patch.object(
                                solr, "_replication_request", return_value="A"
                            ):
                                self.assertDictEqual(
                                    solr.match_index_versions(),
                                    {"A": "a", "errors": [err], "success": False},
                                )

    def test_replication_details(self):
        """
        Test to get the full replication details.
        """
        tempdict1 = {
            "success": "success",
            "errors": "errors",
            "warnings": "warnings",
            "data": "data",
        }

        tempdict2 = {
            "success": None,
            "errors": "errors",
            "warnings": "warnings",
            "data": "data",
        }

        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(solr, "_get_none_or_value", return_value=True):
                with patch.object(
                    solr, "_replication_request", side_effect=[tempdict2, tempdict1]
                ):

                    self.assertDictEqual(solr.replication_details(), tempdict2)

                    with patch.object(
                        solr, "_update_return_dict", return_value=tempdict1
                    ):
                        self.assertDictEqual(solr.replication_details(), tempdict1)

    def test_backup(self):
        """
        Test to tell solr make a backup.
        """
        tempdict = {
            "success": "success",
            "errors": "errors",
            "warnings": "warnings",
            "data": "data",
        }

        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.dict(
                solr.__opts__,
                {
                    "solr.backup_path": MagicMock(return_value="A"),
                    "solr.num_backups": MagicMock(return_value="B"),
                    "solr.cores": MagicMock(return_value=["A"]),
                },
            ):
                with patch.object(os.path, "sep", return_value="B"):
                    with patch.object(
                        solr, "_get_none_or_value", side_effect=[None, True]
                    ):
                        with patch.object(
                            solr, "_check_for_cores", side_effect=[True, False]
                        ):
                            with patch.object(
                                solr, "_replication_request", return_value=tempdict
                            ):
                                with patch.dict(
                                    solr.__opts__,
                                    {"solr.cores": MagicMock(return_value=["A"])},
                                ):
                                    with patch.object(
                                        solr, "_update_return_dict", return_value="A"
                                    ):
                                        self.assertDictEqual(solr.backup(), {"A": "a"})

                                self.assertDictEqual(solr.backup(), tempdict)

    def test_set_is_polling(self):
        """
        Test to prevent the slaves from polling the master for updates.
        """
        tempdict = {
            "success": "success",
            "errors": "errors",
            "warnings": "warnings",
            "data": "data",
        }

        err = 'solr.set_is_polling can only be called by "slave" minions'

        with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
            with patch.object(solr, "_is_master", side_effect=[True, False, False]):
                with patch.object(
                    solr, "_get_none_or_value", side_effect=[None, None, True]
                ):
                    with patch.object(
                        solr, "_check_for_cores", side_effect=[True, False]
                    ):

                        self.assertIsNone(solr.set_is_polling("p"))

                        with patch.dict(
                            solr.__opts__, {"solr.cores": MagicMock(return_value="A")}
                        ):
                            with patch.object(
                                solr, "_update_return_dict", return_value=tempdict
                            ):
                                self.assertDictEqual(
                                    solr.set_is_polling("p"),
                                    {"A": "a", "errors": [err], "success": False},
                                )

                        with patch.object(
                            solr, "_replication_request", return_value="A"
                        ):
                            self.assertEqual(solr.set_is_polling("p"), "A")

    def test_set_replication_enabled(self):
        """
        Test to sets the master to ignore poll requests from the slaves.
        """
        with patch.object(solr, "_is_master", side_effect=[False, True, True, True]):
            with patch.object(
                solr, "_get_none_or_value", side_effect=[None, None, True, True, True]
            ):
                with patch.object(
                    solr, "_get_return_dict", side_effect=[{"A": "a"}, {}]
                ):
                    with patch.object(solr, "_replication_request", return_value="A"):
                        self.assertDictEqual(
                            solr.set_replication_enabled("s"), {"A": "a"}
                        )
                        with patch.object(solr, "_check_for_cores", return_value=True):
                            with patch.dict(
                                solr.__opts__,
                                {"solr.cores": MagicMock(return_value="n")},
                            ):
                                self.assertEqual(solr.set_replication_enabled("s"), {})
                        self.assertEqual(solr.set_replication_enabled("s"), "A")
                        self.assertEqual(solr.set_replication_enabled(False), "A")

    def test_signal(self):
        """
        Test to signals Apache Solr to start, stop, or restart.
        """
        self.assertEqual(
            solr.signal("signal"),
            "signal is an invalid signal. Try: one of: start, stop or restart",
        )

    def test_reload_core(self):
        """
        Test to load a new core from the same configuration as
        an existing registered core.
        """
        error = ['solr.reload_core can only be called by "multi-core" minions']
        with patch.object(
            solr, "_check_for_cores", side_effect=[False, True, True, True]
        ):
            with patch.object(solr, "_get_none_or_value", side_effect=[None, True]):
                with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
                    with patch.object(solr, "_format_url", return_value="A"):
                        with patch.object(solr, "_http_request", return_value="A"):
                            with patch.dict(
                                solr.__opts__,
                                {"solr.cores": MagicMock(return_value="n")},
                            ):
                                self.assertIsNone(solr.reload_core())
                                self.assertDictEqual(
                                    solr.reload_core(),
                                    {"A": "a", "errors": error, "success": False},
                                )
                                self.assertEqual(solr.reload_core(), "A")

    def test_core_status(self):
        """
        Test to get the status for a given core or all cores
        if no core is specified
        """
        error = ['solr.reload_core can only be called by "multi-core" minions']
        with patch.object(
            solr, "_check_for_cores", side_effect=[False, True, True, True]
        ):
            with patch.object(solr, "_get_none_or_value", side_effect=[None, True]):
                with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
                    with patch.object(solr, "_format_url", return_value="A"):
                        with patch.object(solr, "_http_request", return_value="A"):
                            with patch.dict(
                                solr.__opts__,
                                {"solr.cores": MagicMock(return_value="n")},
                            ):
                                self.assertIsNone(solr.core_status())
                                self.assertDictEqual(
                                    solr.core_status(),
                                    {"A": "a", "errors": error, "success": False},
                                )

                                self.assertEqual(solr.core_status(), "A")

    def test_reload_import_config(self):
        """
        Test to re-loads the handler config XML file.
        """
        with patch.object(solr, "_is_master", side_effect=[False, True, True]):
            with patch.object(
                solr, "_get_none_or_value", side_effect=[None, None, None, True, True]
            ):
                with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
                    with patch.object(
                        solr, "_check_for_cores", side_effect=[True, False]
                    ):
                        with patch.object(solr, "_format_url", return_value="A"):
                            with patch.object(solr, "_http_request", return_value="A"):
                                self.assertDictEqual(
                                    solr.reload_import_config("h"), {"A": "a"}
                                )
                                self.assertDictEqual(
                                    solr.reload_import_config("h"), {"A": "a"}
                                )
                                self.assertEqual(solr.reload_import_config("h"), "A")

    def test_abort_import(self):
        """
        Test to aborts an existing import command to the specified handler.
        """
        with patch.object(solr, "_is_master", side_effect=[False, True, True]):
            with patch.object(
                solr, "_get_none_or_value", side_effect=[None, None, None, True, True]
            ):
                with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
                    with patch.object(
                        solr, "_check_for_cores", side_effect=[True, False]
                    ):
                        with patch.object(solr, "_format_url", return_value="A"):
                            with patch.object(solr, "_http_request", return_value="A"):
                                self.assertDictEqual(solr.abort_import("h"), {"A": "a"})
                                self.assertDictEqual(solr.abort_import("h"), {"A": "a"})
                                self.assertEqual(solr.abort_import("h"), "A")

    def test_full_import(self):
        """
        Test to submits an import command to the specified handler using
        specified options.
        """
        with patch(
            "salt.modules.solr._format_url", MagicMock(return_value="A")
        ), patch.object(
            solr, "_is_master", side_effect=[False, True, True, True, True, True]
        ), patch.object(
            solr, "_get_return_dict", return_value={"A": "a"}
        ), patch.object(
            solr, "_get_none_or_value", side_effect=[None, True, True, True, True]
        ), patch.object(
            solr, "_check_for_cores", side_effect=[True, False, False, False, False]
        ), patch.object(
            solr,
            "_pre_index_check",
            side_effect=[{"success": False}, {"success": True}, {"success": True}],
        ), patch.object(
            solr, "_merge_options", side_effect=[{"clean": True}, {"clean": False}]
        ), patch.object(
            solr, "set_replication_enabled", return_value={"success": False}
        ), patch.object(
            solr, "_http_request", return_value="A"
        ):

            self.assertDictEqual(solr.full_import("h"), {"A": "a"})
            self.assertDictEqual(solr.full_import("h"), {"A": "a"})
            self.assertDictEqual(solr.full_import("h"), {"success": False})
            self.assertDictEqual(solr.full_import("h"), {"A": "a"})
            self.assertEqual(solr.full_import("h"), "A")

    def test_delta_import(self):
        """
        Test to submits an import command to the specified handler using
        specified options.
        """
        with patch(
            "salt.modules.solr._format_url", MagicMock(return_value="A")
        ), patch.object(
            solr, "_is_master", side_effect=[False, True, True, True, True]
        ), patch.object(
            solr, "_get_none_or_value", side_effect=[None, True, True, True, True]
        ), patch.object(
            solr, "_get_return_dict", return_value={"A": "a"}
        ), patch.object(
            solr,
            "_pre_index_check",
            side_effect=[
                {"success": False},
                {"success": True},
                {"success": True},
                {"success": True},
            ],
        ), patch.object(
            solr, "_merge_options", side_effect=[{"clean": True}, {"clean": False}]
        ), patch.object(
            solr, "_check_for_cores", side_effect=[True, False]
        ), patch.object(
            solr, "set_replication_enabled", return_value={"success": False}
        ), patch.object(
            solr, "_http_request", return_value="A"
        ):

            self.assertDictEqual(solr.delta_import("h"), {"A": "a"})
            self.assertDictEqual(solr.delta_import("h"), {"success": False})
            self.assertDictEqual(solr.delta_import("h"), {"A": "a"})
            self.assertEqual(solr.delta_import("h"), "A")

    def test_import_status(self):
        """
        Test to submits an import command to the specified handler using
        specified options.
        """
        with patch.object(solr, "_is_master", side_effect=[False, True]):
            with patch.object(solr, "_get_none_or_value", side_effect=[None, True]):
                with patch.object(solr, "_get_return_dict", return_value={"A": "a"}):
                    with patch.object(solr, "_format_url", return_value="A"):
                        with patch.object(solr, "_http_request", return_value="A"):
                            self.assertDictEqual(solr.import_status("h"), {"A": "a"})
                            self.assertEqual(solr.import_status("h"), "A")



import salt.pillar.nodegroups as nodegroups
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

fake_minion_id = "fake_id"
fake_pillar = {}
fake_nodegroups = {
    "groupA": fake_minion_id,
    "groupB": "another_minion_id",
}
fake_opts = {"cache": "localfs", "nodegroups": fake_nodegroups, "id": fake_minion_id}
fake_pillar_name = "fake_pillar_name"


def side_effect(group_sel, t):
    if group_sel.find(fake_minion_id) != -1:
        return {"minions": [fake_minion_id], "missing": []}
    return {"minions": ["another_minion_id"], "missing": []}


class NodegroupsPillarTestCase(TestCase, LoaderModuleMockMixin):
    """
    Tests for salt.pillar.nodegroups
    """

    def setup_loader_modules(self):
        return {nodegroups: {"__opts__": fake_opts}}

    def _runner(self, expected_ret, pillar_name=None):
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(side_effect=side_effect),
        ):
            pillar_name = pillar_name or fake_pillar_name
            actual_ret = nodegroups.ext_pillar(
                fake_minion_id, fake_pillar, pillar_name=pillar_name
            )
            self.assertDictEqual(actual_ret, expected_ret)

    def test_succeeds(self):
        ret = {fake_pillar_name: ["groupA"]}
        self._runner(ret)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import salt.modules.augeas_cfg as augeas_cfg
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

# Make sure augeas python interface is installed
if augeas_cfg.HAS_AUGEAS:
    from augeas import Augeas as _Augeas


@skipIf(augeas_cfg.HAS_AUGEAS is False, "python-augeas is required for this test case")
class AugeasCfgTestCase(TestCase):
    """
    Test cases for salt.modules.augeas_cfg
    """

    # 'execute' function tests: 3
    @skipIf(
        True,
        "Disabled pending https://github.com/hercules-team/python-augeas/issues/30",
    )
    def test_execute(self):
        """
        Test if it execute Augeas commands
        """
        self.assertEqual(augeas_cfg.execute(), {"retval": True})

    def test_execute_io_error(self):
        """
        Test if it execute Augeas commands
        """
        ret = {"error": "Command  is not supported (yet)", "retval": False}
        self.assertEqual(augeas_cfg.execute(None, None, [" "]), ret)

    def test_execute_value_error(self):
        """
        Test if it execute Augeas commands
        """
        ret = {
            "retval": False,
            "error": "Invalid formatted command, see debug log for details: ",
        }
        self.assertEqual(augeas_cfg.execute(None, None, ["set "]), ret)

    # 'get' function tests: 1

    def test_get(self):
        """
        Test if it get a value for a specific augeas path
        """
        mock = MagicMock(side_effect=RuntimeError("error"))
        with patch.object(_Augeas, "match", mock):
            self.assertEqual(augeas_cfg.get("/etc/hosts"), {"error": "error"})

        mock = MagicMock(return_value=True)
        with patch.object(_Augeas, "match", mock):
            self.assertEqual(augeas_cfg.get("/etc/hosts"), {"/etc/hosts": None})

    # 'setvalue' function tests: 4

    def test_setvalue(self):
        """
        Test if it set a value for a specific augeas path
        """
        self.assertEqual(augeas_cfg.setvalue("prefix=/etc/hosts"), {"retval": True})

    def test_setvalue_io_error(self):
        """
        Test if it set a value for a specific augeas path
        """
        mock = MagicMock(side_effect=IOError(""))
        with patch.object(_Augeas, "save", mock):
            self.assertEqual(
                augeas_cfg.setvalue("prefix=/files/etc/"),
                {"retval": False, "error": ""},
            )

    def test_setvalue_uneven_path(self):
        """
        Test if it set a value for a specific augeas path
        """
        mock = MagicMock(side_effect=RuntimeError("error"))
        with patch.object(_Augeas, "match", mock):
            self.assertRaises(
                SaltInvocationError,
                augeas_cfg.setvalue,
                ["/files/etc/hosts/1/canonical", "localhost"],
            )

    def test_setvalue_one_prefix(self):
        """
        Test if it set a value for a specific augeas path
        """
        self.assertRaises(
            SaltInvocationError,
            augeas_cfg.setvalue,
            "prefix=/files",
            "10.18.1.1",
            "prefix=/etc",
            "test",
        )

    # 'match' function tests: 2

    def test_match(self):
        """
        Test if it matches for path expression
        """
        self.assertEqual(augeas_cfg.match("/etc/service", "ssh"), {})

    def test_match_runtime_error(self):
        """
        Test if it matches for path expression
        """
        mock = MagicMock(side_effect=RuntimeError("error"))
        with patch.object(_Augeas, "match", mock):
            self.assertEqual(augeas_cfg.match("/etc/service-name", "ssh"), {})

    # 'remove' function tests: 2

    def test_remove(self):
        """
        Test if it removes for path expression
        """
        self.assertEqual(
            augeas_cfg.remove("/etc/service"), {"count": 0, "retval": True}
        )

    def test_remove_io_runtime_error(self):
        """
        Test if it removes for path expression
        """
        mock = MagicMock(side_effect=RuntimeError("error"))
        with patch.object(_Augeas, "save", mock):
            self.assertEqual(
                augeas_cfg.remove("/etc/service-name"),
                {"count": 0, "error": "error", "retval": False},
            )

    # 'ls' function tests: 1

    def test_ls(self):
        """
        Test if it list the direct children of a node
        """
        self.assertEqual(augeas_cfg.ls("/etc/passwd"), {})

    # 'tree' function tests: 1

    def test_tree(self):
        """
        Test if it returns recursively the complete tree of a node
        """
        self.assertEqual(augeas_cfg.tree("/etc/"), {"/etc": None})



"""
NAPALM Network
==============

Basic methods for interaction with the network device through the virtual proxy 'napalm'.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------

- :mod:`napalm proxy minion <salt.proxy.napalm>`

.. versionadded:: 2016.11.0
.. versionchanged:: 2017.7.0
"""

import datetime
import logging
import time

import salt.utils.files
import salt.utils.napalm
import salt.utils.templates
import salt.utils.versions

log = logging.getLogger(__name__)


try:
    import jxmlease  # pylint: disable=unused-import

    HAS_JXMLEASE = True
except ImportError:
    HAS_JXMLEASE = False

# ----------------------------------------------------------------------------------------------------------------------
# module properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "net"
__proxyenabled__ = ["*"]
__virtual_aliases__ = ("napalm_net",)
# uses NAPALM-based proxy to interact with network devices

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------


def _filter_list(input_list, search_key, search_value):

    """
    Filters a list of dictionary by a set of key-value pair.

    :param input_list:   is a list of dictionaries
    :param search_key:   is the key we are looking for
    :param search_value: is the value we are looking for the key specified in search_key
    :return:             filered list of dictionaries
    """

    output_list = list()

    for dictionary in input_list:
        if dictionary.get(search_key) == search_value:
            output_list.append(dictionary)

    return output_list


def _filter_dict(input_dict, search_key, search_value):

    """
    Filters a dictionary of dictionaries by a key-value pair.

    :param input_dict:    is a dictionary whose values are lists of dictionaries
    :param search_key:    is the key in the leaf dictionaries
    :param search_values: is the value in the leaf dictionaries
    :return:              filtered dictionary
    """

    output_dict = dict()

    for key, key_list in input_dict.items():
        key_list_filtered = _filter_list(key_list, search_key, search_value)
        if key_list_filtered:
            output_dict[key] = key_list_filtered

    return output_dict


def _safe_commit_config(loaded_result, napalm_device):
    _commit = commit(
        inherit_napalm_device=napalm_device
    )  # calls the function commit, defined below
    if not _commit.get("result", False):
        # if unable to commit
        loaded_result["comment"] += (
            _commit["comment"] if _commit.get("comment") else "Unable to commit."
        )
        loaded_result["result"] = False
        # unable to commit, something went wrong
        discarded = _safe_dicard_config(loaded_result, napalm_device)
        if not discarded["result"]:
            return loaded_result
    return _commit


def _safe_dicard_config(loaded_result, napalm_device):
    log.debug("Discarding the config")
    log.debug(loaded_result)
    _discarded = discard_config(inherit_napalm_device=napalm_device)
    if not _discarded.get("result", False):
        loaded_result["comment"] += (
            _discarded["comment"]
            if _discarded.get("comment")
            else "Unable to discard config."
        )
        loaded_result["result"] = False
        # make sure it notifies
        # that something went wrong
        _explicit_close(napalm_device)
        __context__["retcode"] = 1
        return loaded_result
    return _discarded


def _explicit_close(napalm_device):
    """
    Will explicily close the config session with the network device,
    when running in a now-always-alive proxy minion or regular minion.
    This helper must be used in configuration-related functions,
    as the session is preserved and not closed before making any changes.
    """
    if salt.utils.napalm.not_always_alive(__opts__):
        # force closing the configuration session
        # when running in a non-always-alive proxy
        # or regular minion
        try:
            napalm_device["DRIVER"].close()
        except Exception as err:  # pylint: disable=broad-except
            log.error("Unable to close the temp connection with the device:")
            log.error(err)
            log.error("Please report.")


def _config_logic(
    napalm_device,
    loaded_result,
    test=False,
    debug=False,
    replace=False,
    commit_config=True,
    loaded_config=None,
    commit_in=None,
    commit_at=None,
    revert_in=None,
    revert_at=None,
    commit_jid=None,
    **kwargs
):
    """
    Builds the config logic for `load_config` and `load_template` functions.
    """
    # As the Salt logic is built around independent events
    # when it comes to configuration changes in the
    # candidate DB on the network devices, we need to
    # make sure we're using the same session.
    # Hence, we need to pass the same object around.
    # the napalm_device object is inherited from
    # the load_config or load_template functions
    # and forwarded to compare, discard, commit etc.
    # then the decorator will make sure that
    # if not proxy (when the connection is always alive)
    # and the `inherit_napalm_device` is set,
    # `napalm_device` will be overridden.
    # See `salt.utils.napalm.proxy_napalm_wrap` decorator.

    current_jid = kwargs.get("__pub_jid")
    if not current_jid:
        current_jid = "{:%Y%m%d%H%M%S%f}".format(datetime.datetime.now())

    loaded_result["already_configured"] = False

    loaded_result["loaded_config"] = ""
    if debug:
        loaded_result["loaded_config"] = loaded_config

    _compare = compare_config(inherit_napalm_device=napalm_device)
    if _compare.get("result", False):
        loaded_result["diff"] = _compare.get("out")
        loaded_result.pop("out", "")  # not needed
    else:
        loaded_result["diff"] = None
        loaded_result["result"] = False
        loaded_result["comment"] = _compare.get("comment")
        __context__["retcode"] = 1
        return loaded_result

    _loaded_res = loaded_result.get("result", False)
    if not _loaded_res or test:
        # if unable to load the config (errors / warnings)
        # or in testing mode,
        # will discard the config
        if loaded_result["comment"]:
            loaded_result["comment"] += "\n"
        if not len(loaded_result.get("diff", "")) > 0:
            loaded_result["already_configured"] = True
        discarded = _safe_dicard_config(loaded_result, napalm_device)
        if not discarded["result"]:
            return loaded_result
        loaded_result["comment"] += "Configuration discarded."
        # loaded_result['result'] = False not necessary
        # as the result can be true when test=True
        _explicit_close(napalm_device)
        if not loaded_result["result"]:
            __context__["retcode"] = 1
        return loaded_result

    if not test and commit_config:
        # if not in testing mode and trying to commit
        if commit_jid:
            log.info("Committing the JID: %s", str(commit_jid))
            removed = cancel_commit(commit_jid)
            log.debug("Cleaned up the commit from the schedule")
            log.debug(removed["comment"])
        if len(loaded_result.get("diff", "")) > 0:
            # if not testing mode
            # and also the user wants to commit (default)
            # and there are changes to commit
            if commit_in or commit_at:
                commit_time = __utils__["timeutil.get_time_at"](
                    time_in=commit_in, time_at=commit_in
                )
                # schedule job
                scheduled_job_name = "__napalm_commit_{}".format(current_jid)
                temp_file = salt.utils.files.mkstemp()
                with salt.utils.files.fopen(temp_file, "w") as fp_:
                    fp_.write(loaded_config)
                scheduled = __salt__["schedule.add"](
                    scheduled_job_name,
                    function="net.load_config",
                    job_kwargs={
                        "filename": temp_file,
                        "commit_jid": current_jid,
                        "replace": replace,
                    },
                    once=commit_time,
                )
                log.debug("Scheduling job")
                log.debug(scheduled)
                saved = __salt__["schedule.save"]()  # ensure the schedule is
                # persistent cross Minion restart
                discarded = _safe_dicard_config(loaded_result, napalm_device)
                # discard the changes
                if not discarded["result"]:
                    discarded["comment"] += (
                        "Scheduled the job to be executed at {schedule_ts}, "
                        "but was unable to discard the config: \n".format(
                            schedule_ts=commit_time
                        )
                    )
                    return discarded
                loaded_result["comment"] = (
                    "Changes discarded for now, and scheduled commit at:"
                    " {schedule_ts}.\nThe commit ID is: {current_jid}.\nTo discard this"
                    " commit, you can execute: \n\nsalt {min_id} net.cancel_commit"
                    " {current_jid}".format(
                        schedule_ts=commit_time,
                        min_id=__opts__["id"],
                        current_jid=current_jid,
                    )
                )
                loaded_result["commit_id"] = current_jid
                return loaded_result
            log.debug("About to commit:")
            log.debug(loaded_result["diff"])
            if revert_in or revert_at:
                revert_time = __utils__["timeutil.get_time_at"](
                    time_in=revert_in, time_at=revert_at
                )
                if __grains__["os"] == "junos":
                    if not HAS_JXMLEASE:
                        loaded_result["comment"] = (
                            "This feature requires the library jxmlease to be"
                            " installed.\nTo install, please execute: ``pip install"
                            " jxmlease``."
                        )
                        loaded_result["result"] = False
                        return loaded_result
                    timestamp_at = __utils__["timeutil.get_timestamp_at"](
                        time_in=revert_in, time_at=revert_at
                    )
                    minutes = int((timestamp_at - time.time()) / 60)
                    _comm = __salt__["napalm.junos_commit"](confirm=minutes)
                    if not _comm["out"]:
                        # If unable to commit confirm, should try to bail out
                        loaded_result[
                            "comment"
                        ] = "Unable to commit confirm: {}".format(_comm["message"])
                        loaded_result["result"] = False
                        # But before exiting, we must gracefully discard the config
                        discarded = _safe_dicard_config(loaded_result, napalm_device)
                        if not discarded["result"]:
                            return loaded_result
                else:
                    temp_file = salt.utils.files.mkstemp()
                    running_config = __salt__["net.config"](source="running")["out"][
                        "running"
                    ]
                    with salt.utils.files.fopen(temp_file, "w") as fp_:
                        fp_.write(running_config)
                    committed = _safe_commit_config(loaded_result, napalm_device)
                    if not committed["result"]:
                        # If unable to commit, dicard the config (which is
                        # already done by the _safe_commit_config function), and
                        # return with the command and other details.
                        return loaded_result
                    scheduled_job_name = "__napalm_commit_{}".format(current_jid)
                    scheduled = __salt__["schedule.add"](
                        scheduled_job_name,
                        function="net.load_config",
                        job_kwargs={
                            "filename": temp_file,
                            "commit_jid": current_jid,
                            "replace": True,
                        },
                        once=revert_time,
                    )
                    log.debug("Scheduling commit confirmed")
                    log.debug(scheduled)
                    saved = __salt__["schedule.save"]()
                loaded_result["comment"] = (
                    "The commit ID is: {current_jid}.\nThis commit will be reverted at:"
                    " {schedule_ts}, unless confirmed.\nTo confirm the commit and avoid"
                    " reverting, you can execute:\n\nsalt {min_id} net.confirm_commit"
                    " {current_jid}".format(
                        schedule_ts=revert_time,
                        min_id=__opts__["id"],
                        current_jid=current_jid,
                    )
                )
                loaded_result["commit_id"] = current_jid
                return loaded_result
            committed = _safe_commit_config(loaded_result, napalm_device)
            if not committed["result"]:
                return loaded_result
        else:
            # would like to commit, but there's no change
            # need to call discard_config() to release the config DB
            discarded = _safe_dicard_config(loaded_result, napalm_device)
            if not discarded["result"]:
                return loaded_result
            loaded_result["already_configured"] = True
            loaded_result["comment"] = "Already configured."
    _explicit_close(napalm_device)
    if not loaded_result["result"]:
        __context__["retcode"] = 1
    return loaded_result


# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


@salt.utils.napalm.proxy_napalm_wrap
def connected(**kwargs):  # pylint: disable=unused-argument
    """
    Specifies if the connection to the device succeeded.

    CLI Example:

    .. code-block:: bash

        salt '*' net.connected
    """

    return {"out": napalm_device.get("UP", False)}  # pylint: disable=undefined-variable


@salt.utils.napalm.proxy_napalm_wrap
def facts(**kwargs):  # pylint: disable=unused-argument
    """
    Returns characteristics of the network device.
    :return: a dictionary with the following keys:

        * uptime - Uptime of the device in seconds.
        * vendor - Manufacturer of the device.
        * model - Device model.
        * hostname - Hostname of the device
        * fqdn - Fqdn of the device
        * os_version - String with the OS version running on the device.
        * serial_number - Serial number of the device
        * interface_list - List of the interfaces of the device

    CLI Example:

    .. code-block:: bash

        salt '*' net.facts

    Example output:

    .. code-block:: python

        {
            'os_version': '13.3R6.5',
            'uptime': 10117140,
            'interface_list': [
                'lc-0/0/0',
                'pfe-0/0/0',
                'pfh-0/0/0',
                'xe-0/0/0',
                'xe-0/0/1',
                'xe-0/0/2',
                'xe-0/0/3',
                'gr-0/0/10',
                'ip-0/0/10'
            ],
            'vendor': 'Juniper',
            'serial_number': 'JN131356FBFA',
            'model': 'MX480',
            'hostname': 're0.edge05.syd01',
            'fqdn': 're0.edge05.syd01'
        }
    """

    return salt.utils.napalm.call(
        napalm_device, "get_facts", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def environment(**kwargs):  # pylint: disable=unused-argument
    """
    Returns the environment of the device.

    CLI Example:

    .. code-block:: bash

        salt '*' net.environment


    Example output:

    .. code-block:: python

        {
            'fans': {
                'Bottom Rear Fan': {
                    'status': True
                },
                'Bottom Middle Fan': {
                    'status': True
                },
                'Top Middle Fan': {
                    'status': True
                },
                'Bottom Front Fan': {
                    'status': True
                },
                'Top Front Fan': {
                    'status': True
                },
                'Top Rear Fan': {
                    'status': True
                }
            },
            'memory': {
                'available_ram': 16349,
                'used_ram': 4934
            },
            'temperature': {
               'FPC 0 Exhaust A': {
                    'is_alert': False,
                    'temperature': 35.0,
                    'is_critical': False
                }
            },
            'cpu': {
                '1': {
                    '%usage': 19.0
                },
                '0': {
                    '%usage': 35.0
                }
            }
        }
    """

    return salt.utils.napalm.call(
        napalm_device, "get_environment", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def cli(*commands, **kwargs):  # pylint: disable=unused-argument
    """
    Returns a dictionary with the raw output of all commands passed as arguments.

    commands
        List of commands to be executed on the device.

    textfsm_parse: ``False``
        Try parsing the outputs using the TextFSM templates.

        .. versionadded:: 2018.3.0

        .. note::
            This option can be also specified in the minion configuration
            file or pillar as ``napalm_cli_textfsm_parse``.

    textfsm_path
        The path where the TextFSM templates can be found. This option implies
        the usage of the TextFSM index file.
        ``textfsm_path`` can be either absolute path on the server,
        either specified using the following URL mschemes: ``file://``,
        ``salt://``, ``http://``, ``https://``, ``ftp://``,
        ``s3://``, ``swift://``.

        .. versionadded:: 2018.3.0

        .. note::
            This needs to be a directory with a flat structure, having an
            index file (whose name can be specified using the ``index_file`` option)
            and a number of TextFSM templates.

        .. note::
            This option can be also specified in the minion configuration
            file or pillar as ``textfsm_path``.

    textfsm_template
        The path to a certain the TextFSM template.
        This can be specified using the absolute path
        to the file, or using one of the following URL schemes:

        - ``salt://``, to fetch the template from the Salt fileserver.
        - ``http://`` or ``https://``
        - ``ftp://``
        - ``s3://``
        - ``swift://``

        .. versionadded:: 2018.3.0

    textfsm_template_dict
        A dictionary with the mapping between a command
        and the corresponding TextFSM path to use to extract the data.
        The TextFSM paths can be specified as in ``textfsm_template``.

        .. versionadded:: 2018.3.0

        .. note::
            This option can be also specified in the minion configuration
            file or pillar as ``napalm_cli_textfsm_template_dict``.

    platform_grain_name: ``os``
        The name of the grain used to identify the platform name
        in the TextFSM index file. Default: ``os``.

        .. versionadded:: 2018.3.0

        .. note::
            This option can be also specified in the minion configuration
            file or pillar as ``textfsm_platform_grain``.

    platform_column_name: ``Platform``
        The column name used to identify the platform,
        exactly as specified in the TextFSM index file.
        Default: ``Platform``.

        .. versionadded:: 2018.3.0

        .. note::
            This is field is case sensitive, make sure
            to assign the correct value to this option,
            exactly as defined in the index file.

        .. note::
            This option can be also specified in the minion configuration
            file or pillar as ``textfsm_platform_column_name``.

    index_file: ``index``
        The name of the TextFSM index file, under the ``textfsm_path``. Default: ``index``.

        .. versionadded:: 2018.3.0

        .. note::
            This option can be also specified in the minion configuration
            file or pillar as ``textfsm_index_file``.

    saltenv: ``base``
        Salt fileserver environment from which to retrieve the file.
        Ignored if ``textfsm_path`` is not a ``salt://`` URL.

        .. versionadded:: 2018.3.0

    include_empty: ``False``
        Include empty files under the ``textfsm_path``.

        .. versionadded:: 2018.3.0

    include_pat
        Glob or regex to narrow down the files cached from the given path.
        If matching with a regex, the regex must be prefixed with ``E@``,
        otherwise the expression will be interpreted as a glob.

        .. versionadded:: 2018.3.0

    exclude_pat
        Glob or regex to exclude certain files from being cached from the given path.
        If matching with a regex, the regex must be prefixed with ``E@``,
        otherwise the expression will be interpreted as a glob.

        .. versionadded:: 2018.3.0

        .. note::
            If used with ``include_pat``, files matching this pattern will be
            excluded from the subset of files defined by ``include_pat``.

    CLI Example:

    .. code-block:: bash

        salt '*' net.cli "show version" "show chassis fan"

    CLI Example with TextFSM template:

    .. code-block:: bash

        salt '*' net.cli textfsm_parse=True textfsm_path=salt://textfsm/

    Example output:

    .. code-block:: python

        {
            'show version and haiku':  'Hostname: re0.edge01.arn01
                                          Model: mx480
                                          Junos: 13.3R6.5
                                            Help me, Obi-Wan
                                            I just saw Episode Two
                                            You're my only hope
                                         ',
            'show chassis fan' :   'Item                      Status   RPM     Measurement
                                      Top Rear Fan              OK       3840    Spinning at intermediate-speed
                                      Bottom Rear Fan           OK       3840    Spinning at intermediate-speed
                                      Top Middle Fan            OK       3900    Spinning at intermediate-speed
                                      Bottom Middle Fan         OK       3840    Spinning at intermediate-speed
                                      Top Front Fan             OK       3810    Spinning at intermediate-speed
                                      Bottom Front Fan          OK       3840    Spinning at intermediate-speed
                                     '
        }

    Example output with TextFSM parsing:

    .. code-block:: json

        {
          "comment": "",
          "result": true,
          "out": {
            "sh ver": [
              {
                "kernel": "9.1S3.5",
                "documentation": "9.1S3.5",
                "boot": "9.1S3.5",
                "crypto": "9.1S3.5",
                "chassis": "",
                "routing": "9.1S3.5",
                "base": "9.1S3.5",
                "model": "mx960"
              }
            ]
          }
        }
    """
    raw_cli_outputs = salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "cli",
        **{"commands": list(commands)}
    )
    # thus we can display the output as is
    # in case of errors, they'll be caught in the proxy
    if not raw_cli_outputs["result"]:
        # Error -> dispaly the output as-is.
        return raw_cli_outputs
    textfsm_parse = (
        kwargs.get("textfsm_parse")
        or __opts__.get("napalm_cli_textfsm_parse")
        or __pillar__.get("napalm_cli_textfsm_parse", False)
    )
    if not textfsm_parse:
        # No TextFSM parsing required, return raw commands.
        log.debug("No TextFSM parsing requested.")
        return raw_cli_outputs
    if "textfsm.extract" not in __salt__ or "textfsm.index" not in __salt__:
        raw_cli_outputs["comment"] += "Unable to process: is TextFSM installed?"
        log.error(raw_cli_outputs["comment"])
        return raw_cli_outputs
    textfsm_template = kwargs.get("textfsm_template")
    log.debug("textfsm_template: %s", textfsm_template)
    textfsm_path = (
        kwargs.get("textfsm_path")
        or __opts__.get("textfsm_path")
        or __pillar__.get("textfsm_path")
    )
    log.debug("textfsm_path: %s", textfsm_path)
    textfsm_template_dict = (
        kwargs.get("textfsm_template_dict")
        or __opts__.get("napalm_cli_textfsm_template_dict")
        or __pillar__.get("napalm_cli_textfsm_template_dict", {})
    )
    log.debug("TextFSM command-template mapping: %s", textfsm_template_dict)
    index_file = (
        kwargs.get("index_file")
        or __opts__.get("textfsm_index_file")
        or __pillar__.get("textfsm_index_file")
    )
    log.debug("index_file: %s", index_file)
    platform_grain_name = (
        kwargs.get("platform_grain_name")
        or __opts__.get("textfsm_platform_grain")
        or __pillar__.get("textfsm_platform_grain", "os")
    )
    log.debug("platform_grain_name: %s", platform_grain_name)
    platform_column_name = (
        kwargs.get("platform_column_name")
        or __opts__.get("textfsm_platform_column_name")
        or __pillar__.get("textfsm_platform_column_name", "Platform")
    )
    log.debug("platform_column_name: %s", platform_column_name)
    saltenv = kwargs.get("saltenv", "base")
    include_empty = kwargs.get("include_empty", False)
    include_pat = kwargs.get("include_pat")
    exclude_pat = kwargs.get("exclude_pat")
    processed_cli_outputs = {
        "comment": raw_cli_outputs.get("comment", ""),
        "result": raw_cli_outputs["result"],
        "out": {},
    }
    log.debug("Starting to analyse the raw outputs")
    for command in list(commands):
        command_output = raw_cli_outputs["out"][command]
        log.debug("Output from command: %s", command)
        log.debug(command_output)
        processed_command_output = None
        if textfsm_path:
            log.debug("Using the templates under %s", textfsm_path)
            processed_cli_output = __salt__["textfsm.index"](
                command,
                platform_grain_name=platform_grain_name,
                platform_column_name=platform_column_name,
                output=command_output.strip(),
                textfsm_path=textfsm_path,
                saltenv=saltenv,
                include_empty=include_empty,
                include_pat=include_pat,
                exclude_pat=exclude_pat,
            )
            log.debug("Processed CLI output:")
            log.debug(processed_cli_output)
            if not processed_cli_output["result"]:
                log.debug("Apparently this did not work, returning the raw output")
                processed_command_output = command_output
                processed_cli_outputs[
                    "comment"
                ] += "\nUnable to process the output from {}: {}.".format(
                    command, processed_cli_output["comment"]
                )
                log.error(processed_cli_outputs["comment"])
            elif processed_cli_output["out"]:
                log.debug("All good, %s has a nice output!", command)
                processed_command_output = processed_cli_output["out"]
            else:
                comment = """\nProcessing "{}" didn't fail, but didn't return anything either. Dumping raw.""".format(
                    command
                )
                processed_cli_outputs["comment"] += comment
                log.error(comment)
                processed_command_output = command_output
        elif textfsm_template or command in textfsm_template_dict:
            if command in textfsm_template_dict:
                textfsm_template = textfsm_template_dict[command]
            log.debug("Using %s to process the command: %s", textfsm_template, command)
            processed_cli_output = __salt__["textfsm.extract"](
                textfsm_template, raw_text=command_output, saltenv=saltenv
            )
            log.debug("Processed CLI output:")
            log.debug(processed_cli_output)
            if not processed_cli_output["result"]:
                log.debug("Apparently this did not work, returning the raw output")
                processed_command_output = command_output
                processed_cli_outputs[
                    "comment"
                ] += "\nUnable to process the output from {}: {}".format(
                    command, processed_cli_output["comment"]
                )
                log.error(processed_cli_outputs["comment"])
            elif processed_cli_output["out"]:
                log.debug("All good, %s has a nice output!", command)
                processed_command_output = processed_cli_output["out"]
            else:
                log.debug(
                    "Processing %s did not fail, but did not return"
                    " anything either. Dumping raw.",
                    command,
                )
                processed_command_output = command_output
        else:
            log.error("No TextFSM template specified, or no TextFSM path defined")
            processed_command_output = command_output
            processed_cli_outputs[
                "comment"
            ] += "\nUnable to process the output from {}.".format(command)
        processed_cli_outputs["out"][command] = processed_command_output
    processed_cli_outputs["comment"] = processed_cli_outputs["comment"].strip()
    return processed_cli_outputs


@salt.utils.napalm.proxy_napalm_wrap
def traceroute(
    destination, source=None, ttl=None, timeout=None, vrf=None, **kwargs
):  # pylint: disable=unused-argument

    """
    Calls the method traceroute from the NAPALM driver object and returns a dictionary with the result of the traceroute
    command executed on the device.

    destination
        Hostname or address of remote host

    source
        Source address to use in outgoing traceroute packets

    ttl
        IP maximum time-to-live value (or IPv6 maximum hop-limit value)

    timeout
        Number of seconds to wait for response (seconds)

    vrf
        VRF (routing instance) for traceroute attempt

        .. versionadded:: 2016.11.4

    CLI Example:

    .. code-block:: bash

        salt '*' net.traceroute 8.8.8.8
        salt '*' net.traceroute 8.8.8.8 source=127.0.0.1 ttl=5 timeout=1
    """

    return salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "traceroute",
        **{
            "destination": destination,
            "source": source,
            "ttl": ttl,
            "timeout": timeout,
            "vrf": vrf,
        }
    )


@salt.utils.napalm.proxy_napalm_wrap
def ping(
    destination,
    source=None,
    ttl=None,
    timeout=None,
    size=None,
    count=None,
    vrf=None,
    **kwargs
):  # pylint: disable=unused-argument

    """
    Executes a ping on the network device and returns a dictionary as a result.

    destination
        Hostname or IP address of remote host

    source
        Source address of echo request

    ttl
        IP time-to-live value (IPv6 hop-limit value) (1..255 hops)

    timeout
        Maximum wait time after sending final packet (seconds)

    size
        Size of request packets (0..65468 bytes)

    count
        Number of ping requests to send (1..2000000000 packets)

    vrf
        VRF (routing instance) for ping attempt

        .. versionadded:: 2016.11.4

    CLI Example:

    .. code-block:: bash

        salt '*' net.ping 8.8.8.8
        salt '*' net.ping 8.8.8.8 ttl=3 size=65468
        salt '*' net.ping 8.8.8.8 source=127.0.0.1 timeout=1 count=100
    """

    return salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "ping",
        **{
            "destination": destination,
            "source": source,
            "ttl": ttl,
            "timeout": timeout,
            "size": size,
            "count": count,
            "vrf": vrf,
        }
    )


@salt.utils.napalm.proxy_napalm_wrap
def arp(
    interface="", ipaddr="", macaddr="", **kwargs
):  # pylint: disable=unused-argument

    """
    NAPALM returns a list of dictionaries with details of the ARP entries.

    :param interface: interface name to filter on
    :param ipaddr: IP address to filter on
    :param macaddr: MAC address to filter on
    :return: List of the entries in the ARP table

    CLI Example:

    .. code-block:: bash

        salt '*' net.arp
        salt '*' net.arp macaddr='5c:5e:ab:da:3c:f0'

    Example output:

    .. code-block:: python

        [
            {
                'interface' : 'MgmtEth0/RSP0/CPU0/0',
                'mac'       : '5c:5e:ab:da:3c:f0',
                'ip'        : '172.17.17.1',
                'age'       : 1454496274.84
            },
            {
                'interface': 'MgmtEth0/RSP0/CPU0/0',
                'mac'       : '66:0e:94:96:e0:ff',
                'ip'        : '172.17.17.2',
                'age'       : 1435641582.49
            }
        ]
    """

    proxy_output = salt.utils.napalm.call(
        napalm_device, "get_arp_table", **{}  # pylint: disable=undefined-variable
    )

    if not proxy_output.get("result"):
        return proxy_output

    arp_table = proxy_output.get("out")

    if interface:
        arp_table = _filter_list(arp_table, "interface", interface)

    if ipaddr:
        arp_table = _filter_list(arp_table, "ip", ipaddr)

    if macaddr:
        arp_table = _filter_list(arp_table, "mac", macaddr)

    proxy_output.update({"out": arp_table})

    return proxy_output


@salt.utils.napalm.proxy_napalm_wrap
def ipaddrs(**kwargs):  # pylint: disable=unused-argument

    """
    Returns IP addresses configured on the device.

    :return:   A dictionary with the IPv4 and IPv6 addresses of the interfaces.
        Returns all configured IP addresses on all interfaces as a dictionary
        of dictionaries.  Keys of the main dictionary represent the name of the
        interface.  Values of the main dictionary represent are dictionaries
        that may consist of two keys 'ipv4' and 'ipv6' (one, both or none)
        which are themselvs dictionaries with the IP addresses as keys.

    CLI Example:

    .. code-block:: bash

        salt '*' net.ipaddrs

    Example output:

    .. code-block:: python

        {
            'FastEthernet8': {
                'ipv4': {
                    '10.66.43.169': {
                        'prefix_length': 22
                    }
                }
            },
            'Loopback555': {
                'ipv4': {
                    '192.168.1.1': {
                        'prefix_length': 24
                    }
                },
                'ipv6': {
                    '1::1': {
                        'prefix_length': 64
                    },
                    '2001:DB8:1::1': {
                        'prefix_length': 64
                    },
                    'FE80::3': {
                        'prefix_length': 'N/A'
                    }
                }
            }
        }
    """

    return salt.utils.napalm.call(
        napalm_device, "get_interfaces_ip", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def interfaces(**kwargs):  # pylint: disable=unused-argument

    """
    Returns details of the interfaces on the device.

    :return: Returns a dictionary of dictionaries. The keys for the first
        dictionary will be the interfaces in the devices.

    CLI Example:

    .. code-block:: bash

        salt '*' net.interfaces

    Example output:

    .. code-block:: python

        {
            'Management1': {
                'is_up': False,
                'is_enabled': False,
                'description': '',
                'last_flapped': -1,
                'speed': 1000,
                'mac_address': 'dead:beef:dead',
            },
            'Ethernet1':{
                'is_up': True,
                'is_enabled': True,
                'description': 'foo',
                'last_flapped': 1429978575.1554043,
                'speed': 1000,
                'mac_address': 'beef:dead:beef',
            }
        }
    """

    return salt.utils.napalm.call(
        napalm_device, "get_interfaces", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def lldp(interface="", **kwargs):  # pylint: disable=unused-argument

    """
    Returns a detailed view of the LLDP neighbors.

    :param interface: interface name to filter on

    :return:          A dictionary with the LLDL neighbors. The keys are the
        interfaces with LLDP activated on.

    CLI Example:

    .. code-block:: bash

        salt '*' net.lldp
        salt '*' net.lldp interface='TenGigE0/0/0/8'

    Example output:

    .. code-block:: python

        {
            'TenGigE0/0/0/8': [
                {
                    'parent_interface': 'Bundle-Ether8',
                    'interface_description': 'TenGigE0/0/0/8',
                    'remote_chassis_id': '8c60.4f69.e96c',
                    'remote_system_name': 'switch',
                    'remote_port': 'Eth2/2/1',
                    'remote_port_description': 'Ethernet2/2/1',
                    'remote_system_description': 'Cisco Nexus Operating System (NX-OS) Software 7.1(0)N1(1a)
                          TAC support: http://www.cisco.com/tac
                          Copyright (c) 2002-2015, Cisco Systems, Inc. All rights reserved.',
                    'remote_system_capab': 'B, R',
                    'remote_system_enable_capab': 'B'
                }
            ]
        }
    """

    proxy_output = salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "get_lldp_neighbors_detail",
        **{}
    )

    if not proxy_output.get("result"):
        return proxy_output

    lldp_neighbors = proxy_output.get("out")

    if interface:
        lldp_neighbors = {interface: lldp_neighbors.get(interface)}

    proxy_output.update({"out": lldp_neighbors})

    return proxy_output


@salt.utils.napalm.proxy_napalm_wrap
def mac(address="", interface="", vlan=0, **kwargs):  # pylint: disable=unused-argument

    """
    Returns the MAC Address Table on the device.

    :param address:   MAC address to filter on
    :param interface: Interface name to filter on
    :param vlan:      VLAN identifier
    :return:          A list of dictionaries representing the entries in the MAC Address Table

    CLI Example:

    .. code-block:: bash

        salt '*' net.mac
        salt '*' net.mac vlan=10

    Example output:

    .. code-block:: python

        [
            {
                'mac'       : '00:1c:58:29:4a:71',
                'interface' : 'xe-3/0/2',
                'static'    : False,
                'active'    : True,
                'moves'     : 1,
                'vlan'      : 10,
                'last_move' : 1454417742.58
            },
            {
                'mac'       : '8c:60:4f:58:e1:c1',
                'interface' : 'xe-1/0/1',
                'static'    : False,
                'active'    : True,
                'moves'     : 2,
                'vlan'      : 42,
                'last_move' : 1453191948.11
            }
        ]
    """

    proxy_output = salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "get_mac_address_table",
        **{}
    )

    if not proxy_output.get("result"):
        # if negative, leave the output unchanged
        return proxy_output

    mac_address_table = proxy_output.get("out")

    if vlan and isinstance(vlan, int):
        mac_address_table = _filter_list(mac_address_table, "vlan", vlan)

    if address:
        mac_address_table = _filter_list(mac_address_table, "mac", address)

    if interface:
        mac_address_table = _filter_list(mac_address_table, "interface", interface)

    proxy_output.update({"out": mac_address_table})

    return proxy_output


@salt.utils.napalm.proxy_napalm_wrap
def config(source=None, **kwargs):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2017.7.0

    Return the whole configuration of the network device. By default, it will
    return all possible configuration sources supported by the network device.
    At most, there will be:

    - running config
    - startup config
    - candidate config

    To return only one of the configurations, you can use the ``source``
    argument.

    source
        Which configuration type you want to display, default is all of them.

        Options:

        - running
        - candidate
        - startup

    :return:
        The object returned is a dictionary with the following keys:

        - running (string): Representation of the native running configuration.
        - candidate (string): Representation of the native candidate configuration.
            If the device doesn't differentiate between running and startup
            configuration this will an empty string.
        - startup (string): Representation of the native startup configuration.
            If the device doesn't differentiate between running and startup
            configuration this will an empty string.

    CLI Example:

    .. code-block:: bash

        salt '*' net.config
        salt '*' net.config source=candidate
    """
    return salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "get_config",
        **{"retrieve": source}
    )


@salt.utils.napalm.proxy_napalm_wrap
def optics(**kwargs):  # pylint: disable=unused-argument
    """
    .. versionadded:: 2017.7.0

    Fetches the power usage on the various transceivers installed
    on the network device (in dBm), and returns a view that conforms with the
    OpenConfig model openconfig-platform-transceiver.yang.

    :return:
        Returns a dictionary where the keys are as listed below:
            * intf_name (unicode)
                * physical_channels
                    * channels (list of dicts)
                        * index (int)
                        * state
                            * input_power
                                * instant (float)
                                * avg (float)
                                * min (float)
                                * max (float)
                            * output_power
                                * instant (float)
                                * avg (float)
                                * min (float)
                                * max (float)
                            * laser_bias_current
                                * instant (float)
                                * avg (float)
                                * min (float)
                                * max (float)

    CLI Example:

    .. code-block:: bash

        salt '*' net.optics
    """
    return salt.utils.napalm.call(
        napalm_device, "get_optics", **{}  # pylint: disable=undefined-variable
    )


# <---- Call NAPALM getters --------------------------------------------------------------------------------------------

# ----- Configuration specific functions ------------------------------------------------------------------------------>


@salt.utils.napalm.proxy_napalm_wrap
def load_config(
    filename=None,
    text=None,
    test=False,
    commit=True,
    debug=False,
    replace=False,
    commit_in=None,
    commit_at=None,
    revert_in=None,
    revert_at=None,
    commit_jid=None,
    inherit_napalm_device=None,
    saltenv="base",
    **kwargs
):  # pylint: disable=unused-argument
    """
    Applies configuration changes on the device. It can be loaded from a file or from inline string.
    If you send both a filename and a string containing the configuration, the file has higher precedence.

    By default this function will commit the changes. If there are no changes, it does not commit and
    the flag ``already_configured`` will be set as ``True`` to point this out.

    To avoid committing the configuration, set the argument ``test`` to ``True`` and will discard (dry run).

    To keep the changes but not commit, set ``commit`` to ``False``.

    To replace the config, set ``replace`` to ``True``.

    filename
        Path to the file containing the desired configuration.
        This can be specified using the absolute path to the file,
        or using one of the following URL schemes:

        - ``salt://``, to fetch the template from the Salt fileserver.
        - ``http://`` or ``https://``
        - ``ftp://``
        - ``s3://``
        - ``swift://``

        .. versionchanged:: 2018.3.0

    text
        String containing the desired configuration.
        This argument is ignored when ``filename`` is specified.

    test: False
        Dry run? If set as ``True``, will apply the config, discard and return the changes. Default: ``False``
        and will commit the changes on the device.

    commit: True
        Commit? Default: ``True``.

    debug: False
        Debug mode. Will insert a new key under the output dictionary, as ``loaded_config`` containing the raw
        configuration loaded on the device.

        .. versionadded:: 2016.11.2

    replace: False
        Load and replace the configuration. Default: ``False``.

        .. versionadded:: 2016.11.2

    commit_in: ``None``
        Commit the changes in a specific number of minutes / hours. Example of
        accepted formats: ``5`` (commit in 5 minutes), ``2m`` (commit in 2
        minutes), ``1h`` (commit the changes in 1 hour)`, ``5h30m`` (commit
        the changes in 5 hours and 30 minutes).

        .. note::
            This feature works on any platforms, as it does not rely on the
            native features of the network operating system.

        .. note::
            After the command is executed and the ``diff`` is not satisfactory,
            or for any other reasons you have to discard the commit, you are
            able to do so using the
            :py:func:`net.cancel_commit <salt.modules.napalm_network.cancel_commit>`
            execution function, using the commit ID returned by this function.

        .. warning::
            Using this feature, Salt will load the exact configuration you
            expect, however the diff may change in time (i.e., if an user
            applies a manual configuration change, or a different process or
            command changes the configuration in the meanwhile).

        .. versionadded:: 2019.2.0

    commit_at: ``None``
        Commit the changes at a specific time. Example of accepted formats:
        ``1am`` (will commit the changes at the next 1AM), ``13:20`` (will
        commit at 13:20), ``1:20am``, etc.

        .. note::
            This feature works on any platforms, as it does not rely on the
            native features of the network operating system.

        .. note::
            After the command is executed and the ``diff`` is not satisfactory,
            or for any other reasons you have to discard the commit, you are
            able to do so using the
            :py:func:`net.cancel_commit <salt.modules.napalm_network.cancel_commit>`
            execution function, using the commit ID returned by this function.

        .. warning::
            Using this feature, Salt will load the exact configuration you
            expect, however the diff may change in time (i.e., if an user
            applies a manual configuration change, or a different process or
            command changes the configuration in the meanwhile).

        .. versionadded:: 2019.2.0

    revert_in: ``None``
        Commit and revert the changes in a specific number of minutes / hours.
        Example of accepted formats: ``5`` (revert in 5 minutes), ``2m`` (revert
        in 2 minutes), ``1h`` (revert the changes in 1 hour)`, ``5h30m`` (revert
        the changes in 5 hours and 30 minutes).

        .. note::
            To confirm the commit, and prevent reverting the changes, you will
            have to execute the
            :mod:`net.confirm_commit <salt.modules.napalm_network.confirm_commit>`
            function, using the commit ID returned by this function.

        .. warning::
            This works on any platform, regardless if they have or don't have
            native capabilities to confirming a commit. However, please be
            *very* cautious when using this feature: on Junos (as it is the only
            NAPALM core platform supporting this natively) it executes a commit
            confirmed as you would do from the command line.
            All the other platforms don't have this capability natively,
            therefore the revert is done via Salt. That means, your device needs
            to be reachable at the moment when Salt will attempt to revert your
            changes. Be cautious when pushing configuration changes that would
            prevent you reach the device.

            Similarly, if an user or a different process apply other
            configuration changes in the meanwhile (between the moment you
            commit and till the changes are reverted), these changes would be
            equally reverted, as Salt cannot be aware of them.

        .. versionadded:: 2019.2.0

    revert_at: ``None``
        Commit and revert the changes at a specific time. Example of accepted
        formats: ``1am`` (will commit and revert the changes at the next 1AM),
        ``13:20`` (will commit and revert at 13:20), ``1:20am``, etc.

        .. note::
            To confirm the commit, and prevent reverting the changes, you will
            have to execute the
            :mod:`net.confirm_commit <salt.modules.napalm_network.confirm_commit>`
            function, using the commit ID returned by this function.

        .. warning::
            This works on any platform, regardless if they have or don't have
            native capabilities to confirming a commit. However, please be
            *very* cautious when using this feature: on Junos (as it is the only
            NAPALM core platform supporting this natively) it executes a commit
            confirmed as you would do from the command line.
            All the other platforms don't have this capability natively,
            therefore the revert is done via Salt. That means, your device needs
            to be reachable at the moment when Salt will attempt to revert your
            changes. Be cautious when pushing configuration changes that would
            prevent you reach the device.

            Similarly, if an user or a different process apply other
            configuration changes in the meanwhile (between the moment you
            commit and till the changes are reverted), these changes would be
            equally reverted, as Salt cannot be aware of them.

        .. versionadded:: 2019.2.0

    saltenv: ``base``
        Specifies the Salt environment name.

        .. versionadded:: 2018.3.0

    :return: a dictionary having the following keys:

    * result (bool): if the config was applied successfully. It is ``False`` only in case of failure. In case \
    there are no changes to be applied and successfully performs all operations it is still ``True`` and so will be \
    the ``already_configured`` flag (example below)
    * comment (str): a message for the user
    * already_configured (bool): flag to check if there were no changes applied
    * loaded_config (str): the configuration loaded on the device. Requires ``debug`` to be set as ``True``
    * diff (str): returns the config changes applied

    CLI Example:

    .. code-block:: bash

        salt '*' net.load_config text='ntp peer 192.168.0.1'
        salt '*' net.load_config filename='/absolute/path/to/your/file'
        salt '*' net.load_config filename='/absolute/path/to/your/file' test=True
        salt '*' net.load_config filename='/absolute/path/to/your/file' commit=False

    Example output:

    .. code-block:: python

        {
            'comment': 'Configuration discarded.',
            'already_configured': False,
            'result': True,
            'diff': '[edit interfaces xe-0/0/5]+   description "Adding a description";'
        }
    """
    fun = "load_merge_candidate"
    if replace:
        fun = "load_replace_candidate"
    if salt.utils.napalm.not_always_alive(__opts__):
        # if a not-always-alive proxy
        # or regular minion
        # do not close the connection after loading the config
        # this will be handled in _config_logic
        # after running the other features:
        # compare_config, discard / commit
        # which have to be over the same session
        napalm_device["CLOSE"] = False  # pylint: disable=undefined-variable
    if filename:
        text = __salt__["cp.get_file_str"](filename, saltenv=saltenv)
        if text is False:
            # When using salt:// or https://, if the resource is not available,
            #   it will either raise an exception, or return False.
            ret = {"result": False, "out": None}
            ret[
                "comment"
            ] = "Unable to read from {}. Please specify a valid file or text.".format(
                filename
            )
            log.error(ret["comment"])
            return ret
        if commit_jid:
            # When the commit_jid argument is passed, it probably is a scheduled
            # commit to be executed, and filename is a temporary file which
            # can be removed after reading it.
            salt.utils.files.safe_rm(filename)
    _loaded = salt.utils.napalm.call(
        napalm_device, fun, **{"config": text}  # pylint: disable=undefined-variable
    )
    return _config_logic(
        napalm_device,  # pylint: disable=undefined-variable
        _loaded,
        test=test,
        debug=debug,
        replace=replace,
        commit_config=commit,
        loaded_config=text,
        commit_at=commit_at,
        commit_in=commit_in,
        revert_in=revert_in,
        revert_at=revert_at,
        commit_jid=commit_jid,
        **kwargs
    )


@salt.utils.napalm.proxy_napalm_wrap
def load_template(
    template_name=None,
    template_source=None,
    context=None,
    defaults=None,
    template_engine="jinja",
    saltenv="base",
    template_hash=None,
    template_hash_name=None,
    skip_verify=False,
    test=False,
    commit=True,
    debug=False,
    replace=False,
    commit_in=None,
    commit_at=None,
    revert_in=None,
    revert_at=None,
    inherit_napalm_device=None,  # pylint: disable=unused-argument
    **template_vars
):
    """
    Renders a configuration template (default: Jinja) and loads the result on the device.

    By default this function will commit the changes. If there are no changes,
    it does not commit, discards he config and the flag ``already_configured``
    will be set as ``True`` to point this out.

    To avoid committing the configuration, set the argument ``test`` to ``True``
    and will discard (dry run).

    To preserve the changes, set ``commit`` to ``False``.
    However, this is recommended to be used only in exceptional cases
    when there are applied few consecutive states
    and/or configuration changes.
    Otherwise the user might forget that the config DB is locked
    and the candidate config buffer is not cleared/merged in the running config.

    To replace the config, set ``replace`` to ``True``.

    template_name
        Identifies path to the template source.
        The template can be either stored on the local machine, either remotely.
        The recommended location is under the ``file_roots``
        as specified in the master config file.
        For example, let's suppose the ``file_roots`` is configured as:

        .. code-block:: yaml

            file_roots:
              base:
                - /etc/salt/states

        Placing the template under ``/etc/salt/states/templates/example.jinja``,
        it can be used as ``salt://templates/example.jinja``.
        Alternatively, for local files, the user can specify the absolute path.
        If remotely, the source can be retrieved via ``http``, ``https`` or ``ftp``.

        Examples:

        - ``salt://my_template.jinja``
        - ``/absolute/path/to/my_template.jinja``
        - ``http://example.com/template.cheetah``
        - ``https:/example.com/template.mako``
        - ``ftp://example.com/template.py``

        .. versionchanged:: 2019.2.0
            This argument can now support a list of templates to be rendered.
            The resulting configuration text is loaded at once, as a single
            configuration chunk.

    template_source: None
        Inline config template to be rendered and loaded on the device.

    template_hash: None
        Hash of the template file. Format: ``{hash_type: 'md5', 'hsum': <md5sum>}``

        .. versionadded:: 2016.11.2

    context: None
        Overrides default context variables passed to the template.

        .. versionadded:: 2019.2.0

    template_hash_name: None
        When ``template_hash`` refers to a remote file,
        this specifies the filename to look for in that file.

        .. versionadded:: 2016.11.2

    saltenv: ``base``
        Specifies the template environment.
        This will influence the relative imports inside the templates.

        .. versionadded:: 2016.11.2

    template_engine: jinja
        The following templates engines are supported:

        - :mod:`cheetah<salt.renderers.cheetah>`
        - :mod:`genshi<salt.renderers.genshi>`
        - :mod:`jinja<salt.renderers.jinja>`
        - :mod:`mako<salt.renderers.mako>`
        - :mod:`py<salt.renderers.py>`
        - :mod:`wempy<salt.renderers.wempy>`

        .. versionadded:: 2016.11.2

    skip_verify: True
        If ``True``, hash verification of remote file sources
        (``http://``, ``https://``, ``ftp://``) will be skipped,
        and the ``source_hash`` argument will be ignored.

        .. versionadded:: 2016.11.2

    test: False
        Dry run? If set to ``True``, will apply the config,
        discard and return the changes.
        Default: ``False`` and will commit the changes on the device.

    commit: True
        Commit? (default: ``True``)

    debug: False
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw result after the template was rendered.

        .. versionadded:: 2016.11.2

    replace: False
        Load and replace the configuration.

        .. versionadded:: 2016.11.2

    commit_in: ``None``
        Commit the changes in a specific number of minutes / hours. Example of
        accepted formats: ``5`` (commit in 5 minutes), ``2m`` (commit in 2
        minutes), ``1h`` (commit the changes in 1 hour)`, ``5h30m`` (commit
        the changes in 5 hours and 30 minutes).

        .. note::
            This feature works on any platforms, as it does not rely on the
            native features of the network operating system.

        .. note::
            After the command is executed and the ``diff`` is not satisfactory,
            or for any other reasons you have to discard the commit, you are
            able to do so using the
            :py:func:`net.cancel_commit <salt.modules.napalm_network.cancel_commit>`
            execution function, using the commit ID returned by this function.

        .. warning::
            Using this feature, Salt will load the exact configuration you
            expect, however the diff may change in time (i.e., if an user
            applies a manual configuration change, or a different process or
            command changes the configuration in the meanwhile).

        .. versionadded:: 2019.2.0

    commit_at: ``None``
        Commit the changes at a specific time. Example of accepted formats:
        ``1am`` (will commit the changes at the next 1AM), ``13:20`` (will
        commit at 13:20), ``1:20am``, etc.

        .. note::
            This feature works on any platforms, as it does not rely on the
            native features of the network operating system.

        .. note::
            After the command is executed and the ``diff`` is not satisfactory,
            or for any other reasons you have to discard the commit, you are
            able to do so using the
            :py:func:`net.cancel_commit <salt.modules.napalm_network.cancel_commit>`
            execution function, using the commit ID returned by this function.

        .. warning::
            Using this feature, Salt will load the exact configuration you
            expect, however the diff may change in time (i.e., if an user
            applies a manual configuration change, or a different process or
            command changes the configuration in the meanwhile).

        .. versionadded:: 2019.2.0

    revert_in: ``None``
        Commit and revert the changes in a specific number of minutes / hours.
        Example of accepted formats: ``5`` (revert in 5 minutes), ``2m`` (revert
        in 2 minutes), ``1h`` (revert the changes in 1 hour)`, ``5h30m`` (revert
        the changes in 5 hours and 30 minutes).

        .. note::
            To confirm the commit, and prevent reverting the changes, you will
            have to execute the
            :mod:`net.confirm_commit <salt.modules.napalm_network.confirm_commit>`
            function, using the commit ID returned by this function.

        .. warning::
            This works on any platform, regardless if they have or don't have
            native capabilities to confirming a commit. However, please be
            *very* cautious when using this feature: on Junos (as it is the only
            NAPALM core platform supporting this natively) it executes a commit
            confirmed as you would do from the command line.
            All the other platforms don't have this capability natively,
            therefore the revert is done via Salt. That means, your device needs
            to be reachable at the moment when Salt will attempt to revert your
            changes. Be cautious when pushing configuration changes that would
            prevent you reach the device.

            Similarly, if an user or a different process apply other
            configuration changes in the meanwhile (between the moment you
            commit and till the changes are reverted), these changes would be
            equally reverted, as Salt cannot be aware of them.

        .. versionadded:: 2019.2.0

    revert_at: ``None``
        Commit and revert the changes at a specific time. Example of accepted
        formats: ``1am`` (will commit and revert the changes at the next 1AM),
        ``13:20`` (will commit and revert at 13:20), ``1:20am``, etc.

        .. note::
            To confirm the commit, and prevent reverting the changes, you will
            have to execute the
            :mod:`net.confirm_commit <salt.modules.napalm_network.confirm_commit>`
            function, using the commit ID returned by this function.

        .. warning::
            This works on any platform, regardless if they have or don't have
            native capabilities to confirming a commit. However, please be
            *very* cautious when using this feature: on Junos (as it is the only
            NAPALM core platform supporting this natively) it executes a commit
            confirmed as you would do from the command line.
            All the other platforms don't have this capability natively,
            therefore the revert is done via Salt. That means, your device needs
            to be reachable at the moment when Salt will attempt to revert your
            changes. Be cautious when pushing configuration changes that would
            prevent you reach the device.

            Similarly, if an user or a different process apply other
            configuration changes in the meanwhile (between the moment you
            commit and till the changes are reverted), these changes would be
            equally reverted, as Salt cannot be aware of them.

        .. versionadded:: 2019.2.0

    defaults: None
        Default variables/context passed to the template.

        .. versionadded:: 2016.11.2

    template_vars
        Dictionary with the arguments/context to be used when the template is rendered.

        .. note::
            Do not explicitly specify this argument. This represents any other
            variable that will be sent to the template rendering system.
            Please see the examples below!

        .. note::
            It is more recommended to use the ``context`` argument to avoid
            conflicts between CLI arguments and template variables.

    :return: a dictionary having the following keys:

    - result (bool): if the config was applied successfully. It is ``False``
      only in case of failure. In case there are no changes to be applied and
      successfully performs all operations it is still ``True`` and so will be
      the ``already_configured`` flag (example below)
    - comment (str): a message for the user
    - already_configured (bool): flag to check if there were no changes applied
    - loaded_config (str): the configuration loaded on the device, after
      rendering the template. Requires ``debug`` to be set as ``True``
    - diff (str): returns the config changes applied

    The template can use variables from the ``grains``, ``pillar`` or ``opts``, for example:

    .. code-block:: jinja

        {% set router_model = grains.get('model') -%}
        {% set router_vendor = grains.get('vendor') -%}
        {% set os_version = grains.get('version') -%}
        {% set hostname = pillar.get('proxy', {}).get('host') -%}
        {% if router_vendor|lower == 'juniper' %}
        system {
            host-name {{hostname}};
        }
        {% elif router_vendor|lower == 'cisco' %}
        hostname {{hostname}}
        {% endif %}

    CLI Examples:

    .. code-block:: bash

        salt '*' net.load_template set_ntp_peers peers=[192.168.0.1]  # uses NAPALM default templates

        # inline template:
        salt -G 'os:junos' net.load_template template_source='system { host-name {{host_name}}; }' \
        host_name='MX480.lab'

        # inline template using grains info:
        salt -G 'os:junos' net.load_template \
        template_source='system { host-name {{grains.model}}.lab; }'
        # if the device is a MX480, the command above will set the hostname as: MX480.lab

        # inline template using pillar data:
        salt -G 'os:junos' net.load_template template_source='system { host-name {{pillar.proxy.host}}; }'

        salt '*' net.load_template https://bit.ly/2OhSgqP hostname=example  # will commit
        salt '*' net.load_template https://bit.ly/2OhSgqP hostname=example test=True  # dry run

        salt '*' net.load_template salt://templates/example.jinja debug=True  # Using the salt:// URI

        # render a mako template:
        salt '*' net.load_template salt://templates/example.mako template_engine=mako debug=True

        # render remote template
        salt -G 'os:junos' net.load_template http://bit.ly/2fReJg7 test=True debug=True peers=['192.168.0.1']
        salt -G 'os:ios' net.load_template http://bit.ly/2gKOj20 test=True debug=True peers=['192.168.0.1']

        # render multiple templates at once
        salt '*' net.load_template "['https://bit.ly/2OhSgqP', 'salt://templates/example.jinja']" context="{'hostname': 'example'}"

    Example output:

    .. code-block:: python

        {
            'comment': '',
            'already_configured': False,
            'result': True,
            'diff': '[edit system]+  host-name edge01.bjm01',
            'loaded_config': 'system { host-name edge01.bjm01; }''
        }
    """
    _rendered = ""
    _loaded = {"result": True, "comment": "", "out": None}
    loaded_config = None
    # prechecks
    if template_engine not in salt.utils.templates.TEMPLATE_REGISTRY:
        _loaded.update(
            {
                "result": False,
                "comment": (
                    "Invalid templating engine! Choose between: {tpl_eng_opts}".format(
                        tpl_eng_opts=", ".join(
                            list(salt.utils.templates.TEMPLATE_REGISTRY.keys())
                        )
                    )
                ),
            }
        )
        return _loaded  # exit

    # to check if will be rendered by salt or NAPALM
    salt_render_prefixes = ("salt://", "http://", "https://", "ftp://")
    salt_render = False
    file_exists = False
    if not isinstance(template_name, (tuple, list)):
        for salt_render_prefix in salt_render_prefixes:
            if not salt_render:
                salt_render = salt_render or template_name.startswith(
                    salt_render_prefix
                )
        file_exists = __salt__["file.file_exists"](template_name)

    if context is None:
        context = {}
    context.update(template_vars)
    # if needed to render the template send as inline arg
    if template_source:
        # render the content
        _rendered = __salt__["file.apply_template_on_contents"](
            contents=template_source,
            template=template_engine,
            context=context,
            defaults=defaults,
            saltenv=saltenv,
        )
        if not isinstance(_rendered, str):
            if "result" in _rendered:
                _loaded["result"] = _rendered["result"]
            else:
                _loaded["result"] = False
            if "comment" in _rendered:
                _loaded["comment"] = _rendered["comment"]
            else:
                _loaded["comment"] = "Error while rendering the template."
            return _loaded
    else:
        # render the file - either local, either remote
        if not isinstance(template_name, (list, tuple)):
            template_name = [template_name]
        if template_hash_name and not isinstance(template_hash_name, (list, tuple)):
            template_hash_name = [template_hash_name]
        elif not template_hash_name:
            template_hash_name = [None] * len(template_name)
        if (
            template_hash
            and isinstance(template_hash, str)
            and not (
                template_hash.startswith("salt://")
                or template_hash.startswith("file://")
            )
        ):
            # If the template hash is passed as string, and it's not a file
            # (starts with the salt:// or file:// URI), then make it a list
            # of 1 element (for the iteration below)
            template_hash = [template_hash]
        elif (
            template_hash
            and isinstance(template_hash, str)
            and (
                template_hash.startswith("salt://")
                or template_hash.startswith("file://")
            )
        ):
            # If the template hash is a file URI, then provide the same value
            # for each of the templates in the list, as probably they all
            # share the same hash file, otherwise the user should provide
            # this as a list
            template_hash = [template_hash] * len(template_name)
        elif not template_hash:
            template_hash = [None] * len(template_name)
        for tpl_index, tpl_name in enumerate(template_name):
            tpl_hash = template_hash[tpl_index]
            tpl_hash_name = template_hash_name[tpl_index]
            _rand_filename = __salt__["random.hash"](tpl_name, "md5")
            _temp_file = __salt__["file.join"]("/tmp", _rand_filename)
            _managed = __salt__["file.get_managed"](
                name=_temp_file,
                source=tpl_name,
                source_hash=tpl_hash,
                source_hash_name=tpl_hash_name,
                user=None,
                group=None,
                mode=None,
                attrs=None,
                template=template_engine,
                context=context,
                defaults=defaults,
                saltenv=saltenv,
                skip_verify=skip_verify,
            )
            if not isinstance(_managed, (list, tuple)) and isinstance(_managed, str):
                _loaded["comment"] += _managed
                _loaded["result"] = False
            elif isinstance(_managed, (list, tuple)) and not len(_managed) > 0:
                _loaded["result"] = False
                _loaded["comment"] += "Error while rendering the template."
            elif isinstance(_managed, (list, tuple)) and not len(_managed[0]) > 0:
                _loaded["result"] = False
                _loaded["comment"] += _managed[-1]  # contains the error message
            if _loaded["result"]:  # all good
                _temp_tpl_file = _managed[0]
                _temp_tpl_file_exists = __salt__["file.file_exists"](_temp_tpl_file)
                if not _temp_tpl_file_exists:
                    _loaded["result"] = False
                    _loaded["comment"] += "Error while rendering the template."
                    return _loaded
                _rendered += __salt__["file.read"](_temp_tpl_file)
                __salt__["file.remove"](_temp_tpl_file)
            else:
                return _loaded  # exit

    loaded_config = _rendered
    if _loaded["result"]:  # all good
        fun = "load_merge_candidate"
        if replace:  # replace requested
            fun = "load_replace_candidate"
        if salt.utils.napalm.not_always_alive(__opts__):
            # if a not-always-alive proxy
            # or regular minion
            # do not close the connection after loading the config
            # this will be handled in _config_logic
            # after running the other features:
            # compare_config, discard / commit
            # which have to be over the same session
            napalm_device["CLOSE"] = False  # pylint: disable=undefined-variable
        _loaded = salt.utils.napalm.call(
            napalm_device,  # pylint: disable=undefined-variable
            fun,
            **{"config": _rendered}
        )
    return _config_logic(
        napalm_device,  # pylint: disable=undefined-variable
        _loaded,
        test=test,
        debug=debug,
        replace=replace,
        commit_config=commit,
        loaded_config=loaded_config,
        commit_at=commit_at,
        commit_in=commit_in,
        revert_in=revert_in,
        revert_at=revert_at,
        **template_vars
    )


@salt.utils.napalm.proxy_napalm_wrap
def commit(inherit_napalm_device=None, **kwargs):  # pylint: disable=unused-argument

    """
    Commits the configuration changes made on the network device.

    CLI Example:

    .. code-block:: bash

        salt '*' net.commit
    """

    return salt.utils.napalm.call(
        napalm_device, "commit_config", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def discard_config(
    inherit_napalm_device=None, **kwargs
):  # pylint: disable=unused-argument

    """
    Discards the changes applied.

    CLI Example:

    .. code-block:: bash

        salt '*' net.discard_config
    """

    return salt.utils.napalm.call(
        napalm_device, "discard_config", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def compare_config(
    inherit_napalm_device=None, **kwargs
):  # pylint: disable=unused-argument

    """
    Returns the difference between the running config and the candidate config.

    CLI Example:

    .. code-block:: bash

        salt '*' net.compare_config
    """

    return salt.utils.napalm.call(
        napalm_device, "compare_config", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def rollback(inherit_napalm_device=None, **kwargs):  # pylint: disable=unused-argument

    """
    Rollbacks the configuration.

    CLI Example:

    .. code-block:: bash

        salt '*' net.rollback
    """

    return salt.utils.napalm.call(
        napalm_device, "rollback", **{}  # pylint: disable=undefined-variable
    )


@salt.utils.napalm.proxy_napalm_wrap
def config_changed(
    inherit_napalm_device=None, **kwargs
):  # pylint: disable=unused-argument

    """
    Will prompt if the configuration has been changed.

    :return: A tuple with a boolean that specifies if the config was changed on the device.\
    And a string that provides more details of the reason why the configuration was not changed.

    CLI Example:

    .. code-block:: bash

        salt '*' net.config_changed
    """

    is_config_changed = False
    reason = ""
    # pylint: disable=undefined-variable
    try_compare = compare_config(inherit_napalm_device=napalm_device)
    # pylint: enable=undefined-variable

    if try_compare.get("result"):
        if try_compare.get("out"):
            is_config_changed = True
        else:
            reason = "Configuration was not changed on the device."
    else:
        reason = try_compare.get("comment")

    return is_config_changed, reason


@salt.utils.napalm.proxy_napalm_wrap
def config_control(
    inherit_napalm_device=None, **kwargs
):  # pylint: disable=unused-argument

    """
    Will check if the configuration was changed.
    If differences found, will try to commit.
    In case commit unsuccessful, will try to rollback.

    :return: A tuple with a boolean that specifies if the config was changed/committed/rollbacked on the device.\
    And a string that provides more details of the reason why the configuration was not committed properly.

    CLI Example:

    .. code-block:: bash

        salt '*' net.config_control
    """

    result = True
    comment = ""

    # pylint: disable=undefined-variable
    changed, not_changed_rsn = config_changed(inherit_napalm_device=napalm_device)
    # pylint: enable=undefined-variable
    if not changed:
        return (changed, not_changed_rsn)

    # config changed, thus let's try to commit
    try_commit = commit()
    if not try_commit.get("result"):
        result = False
        comment = (
            "Unable to commit the changes: {reason}.\nWill try to rollback now!".format(
                reason=try_commit.get("comment")
            )
        )
        try_rollback = rollback()
        if not try_rollback.get("result"):
            comment += "\nCannot rollback! {reason}".format(
                reason=try_rollback.get("comment")
            )

    return result, comment


def cancel_commit(jid):
    """
    .. versionadded:: 2019.2.0

    Cancel a commit scheduled to be executed via the ``commit_in`` and
    ``commit_at`` arguments from the
    :py:func:`net.load_template <salt.modules.napalm_network.load_template>` or
    :py:func:`net.load_config <salt.modules.napalm_network.load_config>`
    execution functions. The commit ID is displayed when the commit is scheduled
    via the functions named above.

    CLI Example:

    .. code-block:: bash

        salt '*' net.cancel_commit 20180726083540640360
    """
    job_name = "__napalm_commit_{}".format(jid)
    removed = __salt__["schedule.delete"](job_name)
    if removed["result"]:
        saved = __salt__["schedule.save"]()
        removed["comment"] = "Commit #{jid} cancelled.".format(jid=jid)
    else:
        removed["comment"] = "Unable to find commit #{jid}.".format(jid=jid)
    return removed


def confirm_commit(jid):
    """
    .. versionadded:: 2019.2.0

    Confirm a commit scheduled to be reverted via the ``revert_in`` and
    ``revert_at``  arguments from the
    :mod:`net.load_template <salt.modules.napalm_network.load_template>` or
    :mod:`net.load_config <salt.modules.napalm_network.load_config>`
    execution functions. The commit ID is displayed when the commit confirmed
    is scheduled via the functions named above.

    CLI Example:

    .. code-block:: bash

        salt '*' net.confirm_commit 20180726083540640360
    """
    if __grains__["os"] == "junos":
        # Confirm the commit, by committing (i.e., invoking the RPC call)
        confirmed = __salt__["napalm.junos_commit"]()
        confirmed["result"] = confirmed.pop("out")
        confirmed["comment"] = confirmed.pop("message")
    else:
        confirmed = cancel_commit(jid)
    if confirmed["result"]:
        confirmed["comment"] = "Commit #{jid} confirmed.".format(jid=jid)
    return confirmed


def save_config(source=None, path=None):
    """
    .. versionadded:: 2019.2.0

    Save the configuration to a file on the local file system.

    source: ``running``
        The configuration source. Choose from: ``running``, ``candidate``,
        ``startup``. Default: ``running``.

    path
        Absolute path to file where to save the configuration.
        To push the files to the Master, use
        :mod:`cp.push <salt.modules.cp.push>` Execution function.

    CLI Example:

    .. code-block:: bash

        salt '*' net.save_config source=running
    """
    if not source:
        source = "running"
    if not path:
        path = salt.utils.files.mkstemp()
    running_config = __salt__["net.config"](source=source)
    if not running_config or not running_config["result"]:
        log.error("Unable to retrieve the config")
        return running_config
    with salt.utils.files.fopen(path, "w") as fh_:
        fh_.write(running_config["out"][source])
    return {
        "result": True,
        "out": path,
        "comment": "{source} config saved to {path}".format(source=source, path=path),
    }


def replace_pattern(
    pattern,
    repl,
    count=0,
    flags=8,
    bufsize=1,
    append_if_not_found=False,
    prepend_if_not_found=False,
    not_found_content=None,
    search_only=False,
    show_changes=True,
    backslash_literal=False,
    source=None,
    path=None,
    test=False,
    replace=True,
    debug=False,
    commit=True,
):
    """
    .. versionadded:: 2019.2.0

    Replace occurrences of a pattern in the configuration source. If
    ``show_changes`` is ``True``, then a diff of what changed will be returned,
    otherwise a ``True`` will be returned when changes are made, and ``False``
    when no changes are made.
    This is a pure Python implementation that wraps Python's :py:func:`~re.sub`.

    pattern
        A regular expression, to be matched using Python's
        :py:func:`~re.search`.

    repl
        The replacement text.

    count: ``0``
        Maximum number of pattern occurrences to be replaced. If count is a
        positive integer ``n``, only ``n`` occurrences will be replaced,
        otherwise all occurrences will be replaced.

    flags (list or int): ``8``
        A list of flags defined in the ``re`` module documentation from the
        Python standard library. Each list item should be a string that will
        correlate to the human-friendly flag name. E.g., ``['IGNORECASE',
        'MULTILINE']``. Optionally, ``flags`` may be an int, with a value
        corresponding to the XOR (``|``) of all the desired flags. Defaults to
        8 (which supports 'MULTILINE').

    bufsize (int or str): ``1``
        How much of the configuration to buffer into memory at once. The
        default value ``1`` processes one line at a time. The special value
        ``file`` may be specified which will read the entire file into memory
        before processing.

    append_if_not_found: ``False``
        If set to ``True``, and pattern is not found, then the content will be
        appended to the file.

    prepend_if_not_found: ``False``
        If set to ``True`` and pattern is not found, then the content will be
        prepended to the file.

    not_found_content
        Content to use for append/prepend if not found. If None (default), uses
        ``repl``. Useful when ``repl`` uses references to group in pattern.

    search_only: ``False``
        If set to true, this no changes will be performed on the file, and this
        function will simply return ``True`` if the pattern was matched, and
        ``False`` if not.

    show_changes: ``True``
        If ``True``, return a diff of changes made. Otherwise, return ``True``
        if changes were made, and ``False`` if not.

    backslash_literal: ``False``
        Interpret backslashes as literal backslashes for the repl and not
        escape characters.  This will help when using append/prepend so that
        the backslashes are not interpreted for the repl on the second run of
        the state.

    source: ``running``
        The configuration source. Choose from: ``running``, ``candidate``, or
        ``startup``. Default: ``running``.

    path
        Save the temporary configuration to a specific path, then read from
        there.

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard and return
        the changes. Default: ``False`` and will commit the changes on the
        device.

    commit: ``True``
        Commit the configuration changes? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key in the output dictionary, as
        ``loaded_config`` containing the raw configuration loaded on the device.

    replace: ``True``
        Load and replace the configuration. Default: ``True``.

    If an equal sign (``=``) appears in an argument to a Salt command it is
    interpreted as a keyword argument in the format ``key=val``. That
    processing can be bypassed in order to pass an equal sign through to the
    remote shell command by manually specifying the kwarg:

    .. code-block:: bash

        salt '*' net.replace_pattern "bind-address\\s*=" "bind-address:"

    CLI Example:

    .. code-block:: bash

        salt '*' net.replace_pattern PREFIX-LIST_NAME new-prefix-list-name
        salt '*' net.replace_pattern bgp-group-name new-bgp-group-name count=1
    """
    config_saved = save_config(source=source, path=path)
    if not config_saved or not config_saved["result"]:
        return config_saved
    path = config_saved["out"]
    replace_pattern = __salt__["file.replace"](
        path,
        pattern,
        repl,
        count=count,
        flags=flags,
        bufsize=bufsize,
        append_if_not_found=append_if_not_found,
        prepend_if_not_found=prepend_if_not_found,
        not_found_content=not_found_content,
        search_only=search_only,
        show_changes=show_changes,
        backslash_literal=backslash_literal,
    )
    with salt.utils.files.fopen(path, "r") as fh_:
        updated_config = fh_.read()
    return __salt__["net.load_config"](
        text=updated_config, test=test, debug=debug, replace=replace, commit=commit
    )


def blockreplace(
    marker_start,
    marker_end,
    content="",
    append_if_not_found=False,
    prepend_if_not_found=False,
    show_changes=True,
    append_newline=False,
    source="running",
    path=None,
    test=False,
    commit=True,
    debug=False,
    replace=True,
):
    """
    .. versionadded:: 2019.2.0

    Replace content of the configuration source, delimited by the line markers.

    A block of content delimited by comments can help you manage several lines
    without worrying about old entries removal.

    marker_start
        The line content identifying a line as the start of the content block.
        Note that the whole line containing this marker will be considered,
        so whitespace or extra content before or after the marker is included
        in final output.

    marker_end
        The line content identifying a line as the end of the content block.
        Note that the whole line containing this marker will be considered,
        so whitespace or extra content before or after the marker is included
        in final output.

    content
        The content to be used between the two lines identified by
        ``marker_start`` and ``marker_stop``.

    append_if_not_found: ``False``
        If markers are not found and set to True then, the markers and content
        will be appended to the file.

    prepend_if_not_found: ``False``
        If markers are not found and set to True then, the markers and content
        will be prepended to the file.

    append_newline: ``False``
        Controls whether or not a newline is appended to the content block.
        If the value of this argument is ``True`` then a newline will be added
        to the content block. If it is ``False``, then a newline will not be
        added to the content block. If it is ``None`` then a newline will only
        be added to the content block if it does not already end in a newline.

    show_changes: ``True``
        Controls how changes are presented. If ``True``, this function will
        return the of the changes made.
        If ``False``, then it will return a boolean (``True`` if any changes
        were made, otherwise False).

    source: ``running``
        The configuration source. Choose from: ``running``, ``candidate``, or
        ``startup``. Default: ``running``.

    path: ``None``
        Save the temporary configuration to a specific path, then read from
        there. This argument is optional, can be used when you prefers a
        particular location of the temporary file.

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard and return
        the changes. Default: ``False`` and will commit the changes on the
        device.

    commit: ``True``
        Commit the configuration changes? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key in the output dictionary, as
        ``loaded_config`` containing the raw configuration loaded on the device.

    replace: ``True``
        Load and replace the configuration. Default: ``True``.

    CLI Example:

    .. code-block:: bash

        salt '*' net.blockreplace 'ntp' 'interface' ''
    """
    config_saved = save_config(source=source, path=path)
    if not config_saved or not config_saved["result"]:
        return config_saved
    path = config_saved["out"]
    replace_pattern = __salt__["file.blockreplace"](
        path,
        marker_start=marker_start,
        marker_end=marker_end,
        content=content,
        append_if_not_found=append_if_not_found,
        prepend_if_not_found=prepend_if_not_found,
        show_changes=show_changes,
        append_newline=append_newline,
    )
    with salt.utils.files.fopen(path, "r") as fh_:
        updated_config = fh_.read()
    return __salt__["net.load_config"](
        text=updated_config, test=test, debug=debug, replace=replace, commit=commit
    )


def patch(
    patchfile,
    options="",
    saltenv="base",
    source_hash=None,
    show_changes=True,
    source="running",
    path=None,
    test=False,
    commit=True,
    debug=False,
    replace=True,
):
    """
    .. versionadded:: 2019.2.0

    Apply a patch to the configuration source, and load the result into the
    running config of the device.

    patchfile
        A patch file to apply to the configuration source.

    options
        Options to pass to patch.

    source_hash
        If the patch file (specified via the ``patchfile`` argument)  is an
        HTTP(S) or FTP URL and the file exists in the minion's file cache, this
        option can be passed to keep the minion from re-downloading the file if
        the cached copy matches the specified hash.

    show_changes: ``True``
        Controls how changes are presented. If ``True``, this function will
        return the of the changes made.
        If ``False``, then it will return a boolean (``True`` if any changes
        were made, otherwise False).

    source: ``running``
        The configuration source. Choose from: ``running``, ``candidate``, or
        ``startup``. Default: ``running``.

    path: ``None``
        Save the temporary configuration to a specific path, then read from
        there. This argument is optional, can the user prefers a particular
        location of the temporary file.

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard and return
        the changes. Default: ``False`` and will commit the changes on the
        device.

    commit: ``True``
        Commit the configuration changes? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key in the output dictionary, as
        ``loaded_config`` containing the raw configuration loaded on the device.

    replace: ``True``
        Load and replace the configuration. Default: ``True``.

    CLI Example:

    .. code-block:: bash

        salt '*' net.patch https://example.com/running_config.patch
    """
    config_saved = save_config(source=source, path=path)
    if not config_saved or not config_saved["result"]:
        return config_saved
    path = config_saved["out"]
    patchfile_cache = __salt__["cp.cache_file"](patchfile)
    if patchfile_cache is False:
        return {
            "out": None,
            "result": False,
            "comment": 'The file "{}" does not exist.'.format(patchfile),
        }
    replace_pattern = __salt__["file.patch"](path, patchfile_cache, options=options)
    with salt.utils.files.fopen(path, "r") as fh_:
        updated_config = fh_.read()
    return __salt__["net.load_config"](
        text=updated_config, test=test, debug=debug, replace=replace, commit=commit
    )


# <---- Configuration specific functions -------------------------------------------------------------------------------



"""
Execute orchestration functions
"""

import logging

import salt.loader
import salt.utils.event
import salt.utils.functools
import salt.utils.jid
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)


def pause(jid, state_id=None, duration=None):
    """
    Set up a state id pause, this instructs a running state to pause at a given
    state id. This needs to pass in the jid of the running state and can
    optionally pass in a duration in seconds.
    """
    minion = salt.minion.MasterMinion(__opts__)
    minion.functions["state.pause"](jid, state_id, duration)


set_pause = salt.utils.functools.alias_function(pause, "set_pause")


def resume(jid, state_id=None):
    """
    Remove a pause from a jid, allowing it to continue
    """
    minion = salt.minion.MasterMinion(__opts__)
    minion.functions["state.resume"](jid, state_id)


rm_pause = salt.utils.functools.alias_function(resume, "rm_pause")


def soft_kill(jid, state_id=None):
    """
    Set up a state run to die before executing the given state id,
    this instructs a running state to safely exit at a given
    state id. This needs to pass in the jid of the running state.
    If a state_id is not passed then the jid referenced will be safely exited
    at the beginning of the next state run.
    """
    minion = salt.minion.MasterMinion(__opts__)
    minion.functions["state.soft_kill"](jid, state_id)


def orchestrate(
    mods,
    saltenv="base",
    test=None,
    exclude=None,
    pillar=None,
    pillarenv=None,
    pillar_enc=None,
    orchestration_jid=None,
):
    """
    .. versionadded:: 0.17.0

    Execute a state run from the master, used as a powerful orchestration
    system.

    .. seealso:: More Orchestrate documentation

        * :ref:`Full Orchestrate Tutorial <orchestrate-runner>`
        * :py:mod:`Docs for the master-side state module <salt.states.saltmod>`

    CLI Examples:

    .. code-block:: bash

        salt-run state.orchestrate webserver
        salt-run state.orchestrate webserver saltenv=dev test=True
        salt-run state.orchestrate webserver saltenv=dev pillarenv=aws

    .. versionchanged:: 2014.1.1

        Runner renamed from ``state.sls`` to ``state.orchestrate``

    .. versionchanged:: 2014.7.0

        Runner uses the pillar variable

    .. versionchanged:: 2017.5

        Runner uses the pillar_enc variable that allows renderers to render the pillar.
        This is usable when supplying the contents of a file as pillar, and the file contains
        gpg-encrypted entries.

    .. seealso:: GPG renderer documentation

    CLI Examples:

    .. code-block:: bash

       salt-run state.orchestrate webserver pillar_enc=gpg pillar="$(cat somefile.json)"

    """
    if pillar is not None and not isinstance(pillar, dict):
        raise SaltInvocationError("Pillar data must be formatted as a dictionary")
    __opts__["file_client"] = "local"
    minion = salt.minion.MasterMinion(__opts__)

    if pillarenv is None and "pillarenv" in __opts__:
        pillarenv = __opts__["pillarenv"]
    if saltenv is None and "saltenv" in __opts__:
        saltenv = __opts__["saltenv"]
    if orchestration_jid is None:
        orchestration_jid = salt.utils.jid.gen_jid(__opts__)

    running = minion.functions["state.sls"](
        mods,
        test,
        exclude,
        pillar=pillar,
        saltenv=saltenv,
        pillarenv=pillarenv,
        pillar_enc=pillar_enc,
        __pub_jid=orchestration_jid,
        orchestration_jid=orchestration_jid,
    )
    ret = {"data": {minion.opts["id"]: running}, "outputter": "highstate"}
    res = __utils__["state.check_result"](ret["data"])
    if res:
        ret["retcode"] = 0
    else:
        ret["retcode"] = 1
    return ret


# Aliases for orchestrate runner
orch = salt.utils.functools.alias_function(orchestrate, "orch")
sls = salt.utils.functools.alias_function(orchestrate, "sls")


def orchestrate_single(fun, name, test=None, queue=False, pillar=None, **kwargs):
    """
    Execute a single state orchestration routine

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt-run state.orchestrate_single fun=salt.wheel name=key.list_all
    """
    if pillar is not None and not isinstance(pillar, dict):
        raise SaltInvocationError("Pillar data must be formatted as a dictionary")
    __opts__["file_client"] = "local"
    minion = salt.minion.MasterMinion(__opts__)
    running = minion.functions["state.single"](
        fun, name, test=None, queue=False, pillar=pillar, **kwargs
    )
    ret = {minion.opts["id"]: running}
    __jid_event__.fire_event({"data": ret, "outputter": "highstate"}, "progress")
    return ret


def orchestrate_high(data, test=None, queue=False, pillar=None, **kwargs):
    """
    Execute a single state orchestration routine

    .. versionadded:: 2015.5.0

    CLI Example:

    .. code-block:: bash

        salt-run state.orchestrate_high '{
            stage_one:
                {salt.state: [{tgt: "db*"}, {sls: postgres_setup}]},
            stage_two:
                {salt.state: [{tgt: "web*"}, {sls: apache_setup}, {
                    require: [{salt: stage_one}],
                }]},
            }'
    """
    if pillar is not None and not isinstance(pillar, dict):
        raise SaltInvocationError("Pillar data must be formatted as a dictionary")
    __opts__["file_client"] = "local"
    minion = salt.minion.MasterMinion(__opts__)
    running = minion.functions["state.high"](
        data, test=None, queue=False, pillar=pillar, **kwargs
    )
    ret = {minion.opts["id"]: running}
    __jid_event__.fire_event({"data": ret, "outputter": "highstate"}, "progress")
    return ret


def orchestrate_show_sls(
    mods,
    saltenv="base",
    test=None,
    queue=False,
    pillar=None,
    pillarenv=None,
    pillar_enc=None,
):
    """
    Display the state data from a specific sls, or list of sls files, after
    being render using the master minion.

    Note, the master minion adds a "_master" suffix to its minion id.

    .. seealso:: The state.show_sls module function

    CLI Example:

    .. code-block:: bash

        salt-run state.orch_show_sls my-orch-formula.my-orch-state 'pillar={ nodegroup: ng1 }'
    """
    if pillar is not None and not isinstance(pillar, dict):
        raise SaltInvocationError("Pillar data must be formatted as a dictionary")

    __opts__["file_client"] = "local"
    minion = salt.minion.MasterMinion(__opts__)
    running = minion.functions["state.show_sls"](
        mods,
        test,
        queue,
        pillar=pillar,
        pillarenv=pillarenv,
        pillar_enc=pillar_enc,
        saltenv=saltenv,
    )

    ret = {minion.opts["id"]: running}
    return ret


orch_show_sls = salt.utils.functools.alias_function(
    orchestrate_show_sls, "orch_show_sls"
)


def event(
    tagmatch="*", count=-1, quiet=False, sock_dir=None, pretty=False, node="master"
):
    r"""
    Watch Salt's event bus and block until the given tag is matched

    .. versionadded:: 2014.7.0
    .. versionchanged:: 2019.2.0
        ``tagmatch`` can now be either a glob or regular expression.

    This is useful for utilizing Salt's event bus from shell scripts or for
    taking simple actions directly from the CLI.

    Enable debug logging to see ignored events.

    :param tagmatch: the event is written to stdout for each tag that matches
        this glob or regular expression.
    :param count: this number is decremented for each event that matches the
        ``tagmatch`` parameter; pass ``-1`` to listen forever.
    :param quiet: do not print to stdout; just block
    :param sock_dir: path to the Salt master's event socket file.
    :param pretty: Output the JSON all on a single line if ``False`` (useful
        for shell tools); pretty-print the JSON output if ``True``.
    :param node: Watch the minion-side or master-side event bus.
        .. versionadded:: 2016.3.0

    CLI Examples:

    .. code-block:: bash

        # Reboot a minion and run highstate when it comes back online
        salt 'jerry' system.reboot && \\
            salt-run state.event 'salt/minion/jerry/start' count=1 quiet=True && \\
            salt 'jerry' state.highstate

        # Reboot multiple minions and run highstate when all are back online
        salt -L 'kevin,stewart,dave' system.reboot && \\
            salt-run state.event 'salt/minion/*/start' count=3 quiet=True && \\
            salt -L 'kevin,stewart,dave' state.highstate

        # Watch the event bus forever in a shell while-loop.
        salt-run state.event | while read -r tag data; do
            echo $tag
            echo $data | jq --color-output .
        done

    .. seealso::

        See :blob:`tests/eventlisten.sh` for an example of usage within a shell
        script.
    """
    statemod = salt.loader.raw_mod(__opts__, "state", None)

    return statemod["state.event"](
        tagmatch=tagmatch,
        count=count,
        quiet=quiet,
        sock_dir=sock_dir,
        pretty=pretty,
        node=node,
    )



from salt.modules import tuned
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class TunedListTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test the tuned.list_() method for different versions of tuned-adm
    """

    def setup_loader_modules(self):
        return {tuned: {}}

    def test_v_241(self):
        """
        Test the list_ function for older tuned-adm (v2.4.1)
        as shipped with CentOS-6
        """
        tuned_list = """Available profiles:
- throughput-performance
- virtual-guest
- latency-performance
- laptop-battery-powersave
- laptop-ac-powersave
- virtual-host
- desktop-powersave
- server-powersave
- spindown-disk
- sap
- enterprise-storage
- default
Current active profile: throughput-performance"""
        mock_cmd = MagicMock(return_value=tuned_list)
        with patch.dict(tuned.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(
                tuned.list_(),
                [
                    "throughput-performance",
                    "virtual-guest",
                    "latency-performance",
                    "laptop-battery-powersave",
                    "laptop-ac-powersave",
                    "virtual-host",
                    "desktop-powersave",
                    "server-powersave",
                    "spindown-disk",
                    "sap",
                    "enterprise-storage",
                    "default",
                ],
            )

    def test_v_271(self):
        """
        Test the list_ function for newer tuned-adm (v2.7.1)
        as shipped with CentOS-7
        """
        tuned_list = """Available profiles:
- balanced                    - General non-specialized tuned profile
- desktop                     - Optmize for the desktop use-case
- latency-performance         - Optimize for deterministic performance
- network-latency             - Optimize for deterministic performance
- network-throughput          - Optimize for streaming network throughput.
- powersave                   - Optimize for low power-consumption
- throughput-performance      - Broadly applicable tuning that provides--
- virtual-guest               - Optimize for running inside a virtual-guest.
- virtual-host                - Optimize for running KVM guests
Current active profile: virtual-guest
"""
        mock_cmd = MagicMock(return_value=tuned_list)
        with patch.dict(tuned.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(
                tuned.list_(),
                [
                    "balanced",
                    "desktop",
                    "latency-performance",
                    "network-latency",
                    "network-throughput",
                    "powersave",
                    "throughput-performance",
                    "virtual-guest",
                    "virtual-host",
                ],
            )

    def test_v_2110_with_warnings(self):
        """
        Test the list_ function for newer tuned-adm (v2.11.0)
        as shipped with CentOS-7.8 when warnings are emitted
        """
        tuned_list = """Available profiles:
- balanced                    - General non-specialized tuned profile
- desktop                     - Optmize for the desktop use-case
- latency-performance         - Optimize for deterministic performance
- network-latency             - Optimize for deterministic performance
- network-throughput          - Optimize for streaming network throughput.
- powersave                   - Optimize for low power-consumption
- throughput-performance      - Broadly applicable tuning that provides--
- virtual-guest               - Optimize for running inside a virtual-guest.
- virtual-host                - Optimize for running KVM guests
Current active profile: virtual-guest

** COLLECTED WARNINGS **
No SMBIOS nor DMI entry point found, sorry.
** END OF WARNINGS **
"""
        mock_cmd = MagicMock(return_value=tuned_list)
        with patch.dict(tuned.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(
                tuned.list_(),
                [
                    "balanced",
                    "desktop",
                    "latency-performance",
                    "network-latency",
                    "network-throughput",
                    "powersave",
                    "throughput-performance",
                    "virtual-guest",
                    "virtual-host",
                ],
            )

    def test_none(self):
        """ """
        ret = {
            "pid": 12345,
            "retcode": 1,
            "stderr": (
                "stderr: Cannot talk to Tuned daemon via DBus. Is Tuned daemon running?"
            ),
            "stdout": "No current active profile.",
        }
        mock_cmd = MagicMock(return_value=ret)
        with patch.dict(tuned.__salt__, {"cmd.run_all": mock_cmd}):
            self.assertEqual(tuned.active(), "none")



"""
This is the a dummy proxy-minion designed for testing the proxy minion subsystem.
"""

import copy
import logging
import os
import pprint
from contextlib import contextmanager

import salt.utils.files
import salt.utils.msgpack

# This must be present or the Salt loader won't load this module
__proxyenabled__ = ["dummy"]

log = logging.getLogger(__file__)


# This does nothing, it's here just as an example and to provide a log
# entry when the module is loaded.
def __virtual__():
    """
    Only return if all the modules are available
    """
    log.debug("dummy proxy __virtual__() called...")
    return True


def _save_state(opts, details):
    _id = __context__["dummy_proxy"]["id"]
    cachefile = os.path.join(opts["cachedir"], "dummy-proxy-{}.cache".format(_id))
    with salt.utils.files.fopen(cachefile, "wb") as pck:
        pck.write(salt.utils.msgpack.packb(details, use_bin_type=True))
    log.warning("Dummy Proxy Saved State(%s):\n%s", cachefile, pprint.pformat(details))


def _load_state(opts):
    _id = __context__["dummy_proxy"]["id"]
    cachefile = os.path.join(opts["cachedir"], "dummy-proxy-{}.cache".format(_id))
    try:
        with salt.utils.files.fopen(cachefile, "rb") as pck:
            state = salt.utils.msgpack.unpackb(pck.read(), raw=False)
    except FileNotFoundError:
        state = _initial_state()
        _save_state(opts, state)
    except Exception as exc:  # pylint: disable=broad-except
        log.exception("Failed to load state: %s", exc, exc_info=True)
        state = _initial_state()
        _save_state(opts, state)
    log.warning("Dummy Proxy Loaded State(%s):\n%s", cachefile, pprint.pformat(state))
    return state


@contextmanager
def _loaded_state(opts):
    state = _load_state(opts)
    original = copy.deepcopy(state)
    try:
        yield state
    finally:
        if state != original:
            _save_state(opts, state)


def _initial_state():
    return {
        "services": {"apache": "running", "ntp": "running", "samba": "stopped"},
        "packages": {
            "coreutils": "1.0",
            "apache": "2.4",
            "tinc": "1.4",
            "redbull": "999.99",
        },
    }


# Every proxy module needs an 'init', though you can
# just put DETAILS['initialized'] = True here if nothing
# else needs to be done.


def init(opts):
    """
    Required.
    Can be used to initialize the server connection.
    """
    __context__["dummy_proxy"] = {"id": opts["id"]}
    log.debug("dummy proxy init() called...")
    with _loaded_state(opts) as state:
        state["initialized"] = True


def initialized():
    """
    Since grains are loaded in many different places and some of those
    places occur before the proxy can be initialized, return whether
    our init() function has been called
    """
    with _loaded_state(__opts__) as state:
        return state.get("initialized", False)


def grains():
    """
    Make up some grains
    """
    with _loaded_state(__opts__) as state:
        if "grains_cache" not in state:
            state["grains_cache"] = {
                "dummy_grain_1": "one",
                "dummy_grain_2": "two",
                "dummy_grain_3": "three",
            }
        return state["grains_cache"]


def grains_refresh():
    """
    Refresh the grains
    """
    with _loaded_state(__opts__) as state:
        if "grains_cache" in state:
            state.pop("grains_cache")
    return grains()


def fns():
    """
    Method called by grains module.
    """
    return {
        "details": (
            "This key is here because a function in "
            "grains/rest_sample.py called fns() here in the proxymodule."
        )
    }


def service_start(name):
    """
    Start a "service" on the dummy server
    """
    with _loaded_state(__opts__) as state:
        state["services"][name] = "running"
    return "running"


def service_stop(name):
    """
    Stop a "service" on the dummy server
    """
    with _loaded_state(__opts__) as state:
        state["services"][name] = "stopped"
    return "stopped"


def service_restart(name):
    """
    Restart a "service" on the REST server
    """
    return True


def service_list():
    """
    List "services" on the REST server
    """
    with _loaded_state(__opts__) as state:
        return list(state["services"])


def service_status(name):
    """
    Check if a service is running on the REST server
    """
    with _loaded_state(__opts__) as state:
        if state["services"][name] == "running":
            return {"comment": "running"}
        else:
            return {"comment": "stopped"}


def package_list():
    """
    List "packages" installed on the REST server
    """
    with _loaded_state(__opts__) as state:
        return state["packages"]


def package_install(name, **kwargs):
    """
    Install a "package" on the REST server
    """
    if kwargs.get("version", False):
        version = kwargs["version"]
    else:
        version = "1.0"
    with _loaded_state(__opts__) as state:
        state["packages"][name] = version
    return {name: version}


def upgrade():
    """
    "Upgrade" packages
    """
    with _loaded_state(__opts__) as state:
        for p in state["packages"]:
            version_float = float(state["packages"][p])
            version_float = version_float + 1.0
            state["packages"][p] = str(version_float)
        return state["packages"]


def uptodate():
    """
    Call the REST endpoint to see if the packages on the "server" are up to date.
    """
    with _loaded_state(__opts__) as state:
        return state["packages"]


def package_remove(name):
    """
    Remove a "package" on the REST server
    """
    __context__["dummy_proxy"]["foo"] = "bar"
    with _loaded_state(__opts__) as state:
        state["packages"].pop(name)
        return state["packages"]


def package_status(name):
    """
    Check the installation status of a package on the REST server
    """
    with _loaded_state(__opts__) as state:
        if name in state["packages"]:
            return {name: state["packages"][name]}


def ping():
    """
    Degenerate ping
    """
    log.debug("dummy proxy returning ping")
    return True


def shutdown(opts):
    """
    For this proxy shutdown is a no-op
    """
    log.debug("dummy proxy shutdown() called...")
    with _loaded_state(__opts__) as state:
        if "filename" in state:
            os.unlink(state["filename"])


def test_from_state():
    """
    Test function so we have something to call from a state
    :return:
    """
    log.debug("test_from_state called")
    return "testvalue"



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)
    :copyright: Copyright 2014 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    tests.support.xmlunit
    ~~~~~~~~~~~~~~~~~~~

    XML Unit Tests
"""
# pylint: disable=wrong-import-order,wrong-import-position


import io
import logging

log = logging.getLogger(__name__)


try:
    import xmlrunner.runner
    import xmlrunner.result

    HAS_XMLRUNNER = True

    class _DuplicateWriter(io.TextIOBase):
        """
        Duplicate output from the first handle to the second handle
        The second handle is expected to be a StringIO and not to block.
        """

        def __init__(self, first, second):
            super().__init__()
            self._first = first
            self._second = second

        def flush(self):
            self._first.flush()
            self._second.flush()

        def writable(self):
            return True

        def writelines(self, lines):
            self._first.writelines(lines)
            self._second.writelines(lines)

        def write(self, b):
            if isinstance(self._first, io.TextIOBase):
                wrote = self._first.write(b)

                if wrote is not None:
                    # expected to always succeed to write
                    self._second.write(b[:wrote])

                return wrote
            else:
                # file-like object in Python2
                # It doesn't return wrote bytes.
                self._first.write(b)
                self._second.write(b)
                return len(b)

        def fileno(self):
            return self._first.fileno()

    xmlrunner.result._DuplicateWriter = _DuplicateWriter

    class _XMLTestResult(xmlrunner.result._XMLTestResult):
        def startTest(self, test):
            log.debug(">>>>> START >>>>> %s", test.id())
            # xmlrunner classes are NOT new-style classes
            xmlrunner.result._XMLTestResult.startTest(self, test)

        def stopTest(self, test):
            log.debug("<<<<< END <<<<<<< %s", test.id())
            # xmlrunner classes are NOT new-style classes
            return xmlrunner.result._XMLTestResult.stopTest(self, test)

    class XMLTestRunner(xmlrunner.runner.XMLTestRunner):
        def _make_result(self):
            return _XMLTestResult(
                self.stream, self.descriptions, self.verbosity, self.elapsed_times
            )

        def run(self, test):
            result = xmlrunner.runner.XMLTestRunner.run(self, test)
            self.stream.writeln("Finished generating XML reports")
            return result


except ImportError:
    HAS_XMLRUNNER = False

    class XMLTestRunner:
        """
        This is a dumb class just so we don't break projects at import time
        """



"""
InfluxDB - A distributed time series database

Module to provide InfluxDB compatibility to Salt (compatible with InfluxDB
version 0.9+)

:depends:    - influxdb Python module (>= 3.0.0)

:configuration: This module accepts connection configuration details either as
    parameters or as configuration settings in /etc/salt/minion on the relevant
    minions::

        influxdb.host: 'localhost'
        influxdb.port: 8086
        influxdb.user: 'root'
        influxdb.password: 'root'

    This data can also be passed into pillar. Options passed into opts will
    overwrite options passed into pillar.

    Most functions in this module allow you to override or provide some or all
    of these settings via keyword arguments::

        salt '*' influxdb.foo_function influxdb_user='influxadmin' influxdb_password='s3cr1t'

    would override ``user`` and ``password`` while still using the defaults for
    ``host`` and ``port``.
"""

import collections
import logging
from collections.abc import Sequence

import salt.utils.json
from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS

try:
    import influxdb

    HAS_INFLUXDB = True
except ImportError:
    HAS_INFLUXDB = False


log = logging.getLogger(__name__)

# name used to refer to this module in __salt__
__virtualname__ = "influxdb"


def __virtual__():
    """
    Only load if influxdb lib is present
    """
    if HAS_INFLUXDB:
        return __virtualname__
    return (
        False,
        "The influxdb execution module could not be loaded:"
        "influxdb library not available.",
    )


def _client(
    influxdb_user=None,
    influxdb_password=None,
    influxdb_host=None,
    influxdb_port=None,
    **client_args
):
    if not influxdb_user:
        influxdb_user = __salt__["config.option"]("influxdb.user", "root")
    if not influxdb_password:
        influxdb_password = __salt__["config.option"]("influxdb.password", "root")
    if not influxdb_host:
        influxdb_host = __salt__["config.option"]("influxdb.host", "localhost")
    if not influxdb_port:
        influxdb_port = __salt__["config.option"]("influxdb.port", 8086)
    for ignore in _STATE_INTERNAL_KEYWORDS:
        if ignore in client_args:
            del client_args[ignore]
    return influxdb.InfluxDBClient(
        host=influxdb_host,
        port=influxdb_port,
        username=influxdb_user,
        password=influxdb_password,
        **client_args
    )


def list_dbs(**client_args):
    """
    List all InfluxDB databases.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.list_dbs
    """
    client = _client(**client_args)

    return client.get_list_database()


def db_exists(name, **client_args):
    """
    Checks if a database exists in InfluxDB.

    name
        Name of the database to check.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.db_exists <name>
    """
    if name in [db["name"] for db in list_dbs(**client_args)]:
        return True

    return False


def create_db(name, **client_args):
    """
    Create a database.

    name
        Name of the database to create.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.create_db <name>
    """
    if db_exists(name, **client_args):
        log.info("DB '%s' already exists", name)
        return False

    client = _client(**client_args)
    client.create_database(name)

    return True


def drop_db(name, **client_args):
    """
    Drop a database.

    name
        Name of the database to drop.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.drop_db <name>
    """
    if not db_exists(name, **client_args):
        log.info("DB '%s' does not exist", name)
        return False

    client = _client(**client_args)
    client.drop_database(name)

    return True


def list_users(**client_args):
    """
    List all users.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.list_users
    """
    client = _client(**client_args)

    return client.get_list_users()


def user_exists(name, **client_args):
    """
    Check if a user exists.

    name
        Name of the user to check.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.user_exists <name>
    """
    if user_info(name, **client_args):
        return True

    return False


def user_info(name, **client_args):
    """
    Get information about given user.

    name
        Name of the user for which to get information.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.user_info <name>
    """
    matching_users = (
        user for user in list_users(**client_args) if user.get("user") == name
    )

    try:
        return next(matching_users)
    except StopIteration:
        pass


def create_user(name, passwd, admin=False, **client_args):
    """
    Create a user.

    name
        Name of the user to create.

    passwd
        Password of the new user.

    admin : False
        Whether the user should have cluster administration
        privileges or not.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.create_user <name> <password>
        salt '*' influxdb.create_user <name> <password> admin=True
    """
    if user_exists(name, **client_args):
        log.info("User '%s' already exists", name)
        return False

    client = _client(**client_args)
    client.create_user(name, passwd, admin)

    return True


def set_user_password(name, passwd, **client_args):
    """
    Change password of a user.

    name
        Name of the user for whom to set the password.

    passwd
        New password of the user.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.set_user_password <name> <password>
    """
    if not user_exists(name, **client_args):
        log.info("User '%s' does not exist", name)
        return False

    client = _client(**client_args)
    client.set_user_password(name, passwd)

    return True


def grant_admin_privileges(name, **client_args):
    """
    Grant cluster administration privileges to a user.

    name
        Name of the user to whom admin privileges will be granted.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.grant_admin_privileges <name>
    """
    client = _client(**client_args)
    client.grant_admin_privileges(name)

    return True


def revoke_admin_privileges(name, **client_args):
    """
    Revoke cluster administration privileges from a user.

    name
        Name of the user from whom admin privileges will be revoked.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.revoke_admin_privileges <name>
    """
    client = _client(**client_args)
    client.revoke_admin_privileges(name)

    return True


def remove_user(name, **client_args):
    """
    Remove a user.

    name
        Name of the user to remove

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.remove_user <name>
    """
    if not user_exists(name, **client_args):
        log.info("User '%s' does not exist", name)
        return False

    client = _client(**client_args)
    client.drop_user(name)

    return True


def get_retention_policy(database, name, **client_args):
    """
    Get an existing retention policy.

    database
        Name of the database for which the retention policy was
        defined.

    name
        Name of the retention policy.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.get_retention_policy metrics default
    """
    client = _client(**client_args)

    try:
        return next(
            p
            for p in client.get_list_retention_policies(database)
            if p.get("name") == name
        )
    except StopIteration:
        return {}


def retention_policy_exists(database, name, **client_args):
    """
    Check if retention policy with given name exists.

    database
        Name of the database for which the retention policy was
        defined.

    name
        Name of the retention policy to check.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.retention_policy_exists metrics default
    """
    if get_retention_policy(database, name, **client_args):
        return True

    return False


def drop_retention_policy(database, name, **client_args):
    """
    Drop a retention policy.

    database
        Name of the database for which the retention policy will be dropped.

    name
        Name of the retention policy to drop.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.drop_retention_policy mydb mypr
    """
    client = _client(**client_args)
    client.drop_retention_policy(name, database)

    return True


def create_retention_policy(
    database, name, duration, replication, default=False, **client_args
):
    """
    Create a retention policy.

    database
        Name of the database for which the retention policy will be created.

    name
        Name of the new retention policy.

    duration
        Duration of the new retention policy.

        Durations such as 1h, 90m, 12h, 7d, and 4w, are all supported and mean
        1 hour, 90 minutes, 12 hours, 7 day, and 4 weeks, respectively. For
        infinite retention – meaning the data will never be deleted – use 'INF'
        for duration. The minimum retention period is 1 hour.

    replication
        Replication factor of the retention policy.

        This determines how many independent copies of each data point are
        stored in a cluster.

    default : False
        Whether or not the policy as default will be set as default.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.create_retention_policy metrics default 1d 1
    """
    client = _client(**client_args)
    client.create_retention_policy(name, duration, replication, database, default)

    return True


def alter_retention_policy(
    database, name, duration, replication, default=False, **client_args
):
    """
    Modify an existing retention policy.

    name
        Name of the retention policy to modify.

    database
        Name of the database for which the retention policy was defined.

    duration
        New duration of given retention policy.

        Durations such as 1h, 90m, 12h, 7d, and 4w, are all supported
        and mean 1 hour, 90 minutes, 12 hours, 7 day, and 4 weeks,
        respectively. For infinite retention – meaning the data will
        never be deleted – use 'INF' for duration.
        The minimum retention period is 1 hour.

    replication
        New replication of given retention policy.

        This determines how many independent copies of each data point are
        stored in a cluster.

    default : False
        Whether or not to set the modified policy as default.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.alter_retention_policy metrics default 1d 1
    """
    client = _client(**client_args)
    client.alter_retention_policy(name, database, duration, replication, default)
    return True


def list_privileges(name, **client_args):
    """
    List privileges from a user.

    name
        Name of the user from whom privileges will be listed.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.list_privileges <name>
    """
    client = _client(**client_args)

    res = {}
    for item in client.get_list_privileges(name):
        res[item["database"]] = item["privilege"].split()[0].lower()
    return res


def grant_privilege(database, privilege, username, **client_args):
    """
    Grant a privilege on a database to a user.

    database
        Name of the database to grant the privilege on.

    privilege
        Privilege to grant. Can be one of 'read', 'write' or 'all'.

    username
        Name of the user to grant the privilege to.
    """
    client = _client(**client_args)
    client.grant_privilege(privilege, database, username)

    return True


def revoke_privilege(database, privilege, username, **client_args):
    """
    Revoke a privilege on a database from a user.

    database
        Name of the database to grant the privilege on.

    privilege
        Privilege to grant. Can be one of 'read', 'write' or 'all'.

    username
        Name of the user to grant the privilege to.
    """
    client = _client(**client_args)
    client.revoke_privilege(privilege, database, username)

    return True


def continuous_query_exists(database, name, **client_args):
    """
    Check if continuous query with given name exists on the database.

    database
        Name of the database for which the continuous query was
        defined.

    name
        Name of the continuous query to check.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.continuous_query_exists metrics default
    """
    if get_continuous_query(database, name, **client_args):
        return True

    return False


def get_continuous_query(database, name, **client_args):
    """
    Get an existing continuous query.

    database
        Name of the database for which the continuous query was
        defined.

    name
        Name of the continuous query to get.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.get_continuous_query mydb cq_month
    """
    client = _client(**client_args)

    try:
        for db, cqs in client.query("SHOW CONTINUOUS QUERIES").items():
            if db[0] == database:
                return next(cq for cq in cqs if cq.get("name") == name)
    except StopIteration:
        return {}
    return {}


def create_continuous_query(
    database, name, query, resample_time=None, coverage_period=None, **client_args
):
    """
    Create a continuous query.

    database
        Name of the database for which the continuous query will be
        created on.

    name
        Name of the continuous query to create.

    query
        The continuous query string.

    resample_time : None
        Duration between continuous query resampling.

    coverage_period : None
        Duration specifying time period per sample.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.create_continuous_query mydb cq_month 'SELECT mean(*) INTO mydb.a_month.:MEASUREMENT FROM mydb.a_week./.*/ GROUP BY time(5m), *'"""
    client = _client(**client_args)
    full_query = "CREATE CONTINUOUS QUERY {name} ON {database}"
    if resample_time:
        full_query += " RESAMPLE EVERY {resample_time}"
    if coverage_period:
        full_query += " FOR {coverage_period}"
    full_query += " BEGIN {query} END"
    query = full_query.format(
        name=name,
        database=database,
        query=query,
        resample_time=resample_time,
        coverage_period=coverage_period,
    )
    client.query(query)
    return True


def drop_continuous_query(database, name, **client_args):
    """
    Drop a continuous query.

    database
        Name of the database for which the continuous query will
        be drop from.

    name
        Name of the continuous query to drop.

    CLI Example:

    .. code-block:: bash

        salt '*' influxdb.drop_continuous_query mydb my_cq
    """
    client = _client(**client_args)

    query = "DROP CONTINUOUS QUERY {} ON {}".format(name, database)
    client.query(query)
    return True


def _pull_query_results(resultset):
    """
    Parses a ResultSet returned from InfluxDB into a dictionary of results,
    grouped by series names and optional JSON-encoded grouping tags.
    """
    _results = collections.defaultdict(lambda: {})
    for _header, _values in resultset.items():
        _header, _group_tags = _header
        if _group_tags:
            _results[_header][salt.utils.json.dumps(_group_tags)] = [
                _value for _value in _values
            ]
        else:
            _results[_header] = [_value for _value in _values]
    return dict(sorted(_results.items()))


def query(database, query, **client_args):
    """
    Execute a query.

    database
        Name of the database to query on.

    query
        InfluxQL query string.
    """
    client = _client(**client_args)
    _result = client.query(query, database=database)

    if isinstance(_result, Sequence):
        return [
            _pull_query_results(_query_result)
            for _query_result in _result
            if _query_result
        ]
    return [_pull_query_results(_result) if _result else {}]
